# -*- coding: utf-8 -*-
"""The CPIO file entry implementation."""

import stat

from dfdatetime import posix_time as dfdatetime_posix_time

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import cpio_path_spec
from dfvfs.vfs import attribute
from dfvfs.vfs import cpio_directory
from dfvfs.vfs import file_entry


class CPIOFileEntry(file_entry.FileEntry):
  """File system file entry that uses CPIOArchiveFile."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_CPIO

  def __init__(
      self, resolver_context, file_system, path_spec,
      cpio_archive_file_entry=None, is_root=False, is_virtual=False):
    """Initializes a file entry object.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      cpio_archive_file_entry (Optional[CPIOArchiveFileEntry]): CPIO archive
          file entry.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file
          entry emulated by the corresponding file system.

    Raises:
      BackEndError: when the CPIO archive file entry is missing in
          a non-virtual file entry.
    """
    if not is_virtual and cpio_archive_file_entry is None:
      cpio_archive_file_entry = file_system.GetCPIOArchiveFileEntryByPathSpec(
          path_spec)
    if not is_virtual and cpio_archive_file_entry is None:
      raise errors.BackEndError(
          'Missing CPIO archive file entry in non-virtual file entry.')

    super(CPIOFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._cpio_archive_file_entry = cpio_archive_file_entry

    # The stat info member st_mode can have multiple types e.g.
    # LINK and DIRECTORY in case of a symbolic link to a directory
    # dfVFS currently only supports one type so we need to check
    # for LINK first.
    mode = getattr(cpio_archive_file_entry, 'mode', 0)
    if stat.S_ISLNK(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_LINK
    # The root file entry is virtual and should have type directory.
    elif is_virtual or stat.S_ISDIR(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_DIRECTORY
    elif stat.S_ISREG(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_FILE
    elif stat.S_ISCHR(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_CHARACTER_DEVICE
    elif stat.S_ISBLK(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_BLOCK_DEVICE
    elif stat.S_ISFIFO(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_PIPE
    elif stat.S_ISSOCK(mode):
      self.entry_type = definitions.FILE_ENTRY_TYPE_SOCKET

  def _GetDirectory(self):
    """Retrieves a directory.

    Returns:
      CPIODirectory: a directory.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_DIRECTORY:
      return None

    return cpio_directory.CPIODirectory(self._file_system, self.path_spec)

  def _GetLink(self):
    """Retrieves the link.

    Returns:
      str: full path of the linked file entry.
    """
    if self._link is None:
      cpio_archive_file = self._file_system.GetCPIOArchiveFile()
      link_data = cpio_archive_file.ReadDataAtOffset(
          self._cpio_archive_file_entry.data_offset,
          self._cpio_archive_file_entry.data_size)

      self._link = link_data.decode(cpio_archive_file.encoding)

    return self._link

  def _GetStatAttribute(self):
    """Retrieves a stat attribute.

    Returns:
      StatAttribute: a stat attribute or None if not available.
    """
    mode = getattr(self._cpio_archive_file_entry, 'mode', 0)

    stat_attribute = attribute.StatAttribute()
    stat_attribute.group_identifier = getattr(
        self._cpio_archive_file_entry, 'group_identifier', None)
    stat_attribute.inode_number = getattr(
        self._cpio_archive_file_entry, 'inode_number', None)
    stat_attribute.mode = stat.S_IMODE(mode)
    stat_attribute.number_of_links = getattr(
        self._cpio_archive_file_entry, 'number_of_links', None)
    stat_attribute.owner_identifier = getattr(
        self._cpio_archive_file_entry, 'user_identifier', None)
    stat_attribute.size = getattr(
        self._cpio_archive_file_entry, 'data_size', None)
    stat_attribute.type = self.entry_type

    return stat_attribute

  def _GetSubFileEntries(self):
    """Retrieves sub file entries.

    Yields:
      CPIOFileEntry: a sub file entry.
    """
    if self._directory is None:
      self._directory = self._GetDirectory()

    if self._directory:
      for path_spec in self._directory.entries:
        cpio_archive_file_entry = (
            self._file_system.GetCPIOArchiveFileEntryByPathSpec(path_spec))
        is_virtual = not bool(cpio_archive_file_entry)

        yield CPIOFileEntry(
            self._resolver_context, self._file_system, path_spec,
            is_virtual=is_virtual)

  @property
  def name(self):
    """str: name of the file entry, which does not include the full path."""
    # Note that the root file entry is virtual and has no
    # cpio_archive_file_entry.
    if self._cpio_archive_file_entry is None:
      return ''

    return self._file_system.BasenamePath(self._cpio_archive_file_entry.path)

  @property
  def modification_time(self):
    """dfdatetime.DateTimeValues: modification time or None if not available."""
    timestamp = getattr(
        self._cpio_archive_file_entry, 'modification_time', None)
    if timestamp is None:
      return None
    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    return getattr(self._cpio_archive_file_entry, 'data_size', None)

  def GetCPIOArchiveFileEntry(self):
    """Retrieves the CPIO archive file entry object.

    Returns:
      CPIOArchiveFileEntry: CPIO archive file entry.

    Raises:
      PathSpecError: if the path specification is incorrect.
    """
    return self._cpio_archive_file_entry

  def GetParentFileEntry(self):
    """Retrieves the parent file entry.

    Returns:
      CPIOFileEntry: parent file entry or None if not available.
    """
    location = getattr(self.path_spec, 'location', None)
    if location is None:
      return None

    parent_location = self._file_system.DirnamePath(location)
    if parent_location is None:
      return None

    if parent_location == '':
      parent_location = self._file_system.PATH_SEPARATOR
      is_root = True
      is_virtual = True
    else:
      is_root = False
      is_virtual = False

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    path_spec = cpio_path_spec.CPIOPathSpec(
        location=parent_location, parent=parent_path_spec)
    return CPIOFileEntry(
        self._resolver_context, self._file_system, path_spec,
        is_root=is_root, is_virtual=is_virtual)



# -*- coding: utf-8 -*-
"""Helper function for LUKS Drive Encryption support."""


def LUKSDEOpenVolume(luksde_volume, path_spec, file_object, key_chain):
  """Opens the LUKSDE volume using the path specification.

  Args:
    luksde_volume (pyluksde.volume): LUKSDE volume.
    path_spec (PathSpec): path specification.
    file_object (FileIO): file-like object.
    key_chain (KeyChain): key chain.
  """
  password = key_chain.GetCredential(path_spec, 'password')
  if password:
    luksde_volume.set_password(password)

  luksde_volume.open_file_object(file_object)


def LUKSDEUnlockVolume(luksde_volume, path_spec, key_chain):
  """Unlocks a LUKSDE volume using the path specification.

  Args:
    luksde_volume (pyluksde.volume): LUKSDE volume.
    path_spec (PathSpec): path specification.
    key_chain (KeyChain): key chain.

  Returns:
    bool: True if the volume is unlocked, False otherwise.
  """
  is_locked = luksde_volume.is_locked()
  if is_locked:
    password = key_chain.GetCredential(path_spec, 'password')
    if password:
      luksde_volume.set_password(password)

    is_locked = not luksde_volume.unlock()

  return not is_locked



# -*- coding: utf-8 -*-
"""The GUID Partition Table (GPT) directory implementation."""

from dfvfs.path import gpt_path_spec
from dfvfs.vfs import directory


class GPTDirectory(directory.Directory):
  """File system directory that uses pyvsgpt."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      GPTPathSpec: a path specification.
    """
    entry_index = getattr(self.path_spec, 'entry_index', None)
    location = getattr(self.path_spec, 'location', None)

    # Only the virtual root file has directory entries.
    if (entry_index is None and location is not None and
        location == self._file_system.LOCATION_ROOT):
      vsgpt_volume = self._file_system.GetGPTVolume()

      for partition in vsgpt_volume.partitions:
        gpt_entry_index = partition.entry_index + 1
        yield gpt_path_spec.GPTPathSpec(
            entry_index=entry_index, location=f'/p{gpt_entry_index:d}',
            parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""The FAT file system implementation."""

import pyfsfat

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import fat_path_spec
from dfvfs.resolver import resolver
from dfvfs.vfs import fat_file_entry
from dfvfs.vfs import file_system


class FATFileSystem(file_system.FileSystem):
  """File system that uses pyfsfat."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_FAT

  LOCATION_ROOT = '\\'
  PATH_SEPARATOR = '\\'

  def __init__(self, resolver_context, path_spec):
    """Initializes an FAT file system.

    Args:
      resolver_context (Context): resolver context.
      path_spec (PathSpec): a path specification.
    """
    super(FATFileSystem, self).__init__(resolver_context, path_spec)
    self._file_object = None
    self._fsfat_volume = None
    self._root_directory_identifier = None

  def _Close(self):
    """Closes the file system.

    Raises:
      IOError: if the close failed.
    """
    self._fsfat_volume = None
    self._file_object = None

  def _Open(self, mode='rb'):
    """Opens the file system defined by path specification.

    Args:
      mode (Optional[str]): file access mode.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file system object could not be opened.
      PathSpecError: if the path specification is incorrect.
      ValueError: if the path specification is invalid.
    """
    if not self._path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification without parent.')

    file_object = resolver.Resolver.OpenFileObject(
        self._path_spec.parent, resolver_context=self._resolver_context)

    fsfat_volume = pyfsfat.volume()
    fsfat_volume.open_file_object(file_object)
    fsfat_root_directory = fsfat_volume.get_root_directory()

    self._file_object = file_object
    self._fsfat_volume = fsfat_volume
    self._root_directory_identifier = fsfat_root_directory.identifier

  def FileEntryExistsByPathSpec(self, path_spec):
    """Determines if a file entry for a path specification exists.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      bool: True if the file entry exists.

    Raises:
      BackEndError: if the file entry cannot be opened.
    """
    # Opening a file by identifier is faster than opening a file by location.
    fsfat_file_entry = None
    location = getattr(path_spec, 'location', None)
    identifier = getattr(path_spec, 'identifier', None)

    try:
      if identifier is not None:
        fsfat_file_entry = self._fsfat_volume.get_file_entry_by_identifier(
            identifier)
      elif location is not None:
        fsfat_file_entry = self._fsfat_volume.get_file_entry_by_path(location)

    except IOError as exception:
      raise errors.BackEndError(exception)

    return fsfat_file_entry is not None

  def GetFileEntryByPathSpec(self, path_spec):
    """Retrieves a file entry for a path specification.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      FATFileEntry: file entry or None if not available.

    Raises:
      BackEndError: if the file entry cannot be opened.
    """
    # Opening a file by identifier is faster than opening a file by location.
    fsfat_file_entry = None
    location = getattr(path_spec, 'location', None)
    identifier = getattr(path_spec, 'identifier', None)

    if (location == self.LOCATION_ROOT or
        identifier == self._root_directory_identifier):
      fsfat_file_entry = self._fsfat_volume.get_root_directory()
      return fat_file_entry.FATFileEntry(
          self._resolver_context, self, path_spec,
          fsfat_file_entry=fsfat_file_entry, is_root=True)

    try:
      if identifier is not None:
        fsfat_file_entry = self._fsfat_volume.get_file_entry_by_identifier(
            identifier)
      elif location is not None:
        fsfat_file_entry = self._fsfat_volume.get_file_entry_by_path(location)

    except IOError as exception:
      raise errors.BackEndError(exception)

    if fsfat_file_entry is None:
      return None

    return fat_file_entry.FATFileEntry(
        self._resolver_context, self, path_spec,
        fsfat_file_entry=fsfat_file_entry)

  def GetFATFileEntryByPathSpec(self, path_spec):
    """Retrieves the FAT file entry for a path specification.

    Args:
      path_spec (PathSpec): a path specification.

    Returns:
      pyfsfat.file_entry: file entry.

    Raises:
      PathSpecError: if the path specification is missing location and
          identifier.
    """
    # Opening a file by identifier is faster than opening a file by location.
    location = getattr(path_spec, 'location', None)
    identifier = getattr(path_spec, 'identifier', None)

    if identifier is not None:
      fsfat_file_entry = self._fsfat_volume.get_file_entry_by_identifier(
          identifier)
    elif location is not None:
      fsfat_file_entry = self._fsfat_volume.get_file_entry_by_path(location)
    else:
      raise errors.PathSpecError(
          'Path specification missing location and identifier.')

    return fsfat_file_entry

  def GetFATVolume(self):
    """Retrieves the FAT volume.

    Returns:
      pyfsfat.volume: a FAT volume.
    """
    return self._fsfat_volume

  def GetRootFileEntry(self):
    """Retrieves the root file entry.

    Returns:
      FATFileEntry: file entry.
    """
    path_spec = fat_path_spec.FATPathSpec(
        location=self.LOCATION_ROOT,
        identifier=self._root_directory_identifier,
        parent=self._path_spec.parent)
    return self.GetFileEntryByPathSpec(path_spec)



# -*- coding: utf-8 -*-
"""The RC4 decrypter implementation."""

import pyfcrypto

from dfvfs.encryption import decrypter
from dfvfs.encryption import manager
from dfvfs.lib import definitions


class RC4Decrypter(decrypter.Decrypter):
  """RC4 decrypter using pyfcrypto."""

  ENCRYPTION_METHOD = definitions.ENCRYPTION_METHOD_RC4

  def __init__(self, key=None, **kwargs):
    """Initializes a decrypter.

    Args:
      key (Optional[bytes]): key.
      kwargs (dict): keyword arguments depending on the decrypter.

    Raises:
      ValueError: when key is not set.
    """
    if not key:
      raise ValueError('Missing key.')

    super(RC4Decrypter, self).__init__(**kwargs)
    self._rc4_context = pyfcrypto.rc4_context()

    self._rc4_context.set_key(key)

  # pylint: disable=unused-argument
  def Decrypt(self, encrypted_data, finalize=False):
    """Decrypts the encrypted data.

    Args:
      encrypted_data (bytes): encrypted data.
      finalize (Optional[bool]): True if the end of data has been reached and
          the cipher context should be finalized.

    Returns:
      tuple[bytes,bytes]: decrypted data and remaining encrypted data.
    """
    decrypted_data = pyfcrypto.crypt_rc4(self._rc4_context, encrypted_data)
    return decrypted_data, b''


manager.EncryptionManager.RegisterDecrypter(RC4Decrypter)



# -*- coding: utf-8 -*-
"""The SleuthKit (TSK) directory implementation."""

import pytsk3

from dfvfs.lib import errors
from dfvfs.path import tsk_path_spec
from dfvfs.vfs import directory


class TSKDirectory(directory.Directory):
  """File system directory that uses pytsk3."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      TSKPathSpec: a path specification.

    Raises:
      BackEndError: if pytsk3 cannot open the directory.
    """
    # Opening a file by inode number is faster than opening a file
    # by location.
    inode = getattr(self.path_spec, 'inode', None)
    location = getattr(self.path_spec, 'location', None)

    fs_info = self._file_system.GetFsInfo()
    tsk_directory = None

    try:
      if inode is not None:
        tsk_directory = fs_info.open_dir(inode=inode)
      elif location is not None:
        tsk_directory = fs_info.open_dir(path=location)

    except IOError as exception:
      raise errors.BackEndError(
          f'Unable to open directory with error: {exception!s}')

    if tsk_directory:
      for tsk_directory_entry in tsk_directory:
        # Note that because pytsk3.Directory does not explicitly define info
        # we need to check if the attribute exists and has a value other
        # than None.
        if getattr(tsk_directory_entry, 'info', None) is None:
          continue

        # Note that because pytsk3.TSK_FS_FILE does not explicitly define
        # fs_info we need to check if the attribute exists and has a value
        # other than None.
        if getattr(tsk_directory_entry.info, 'fs_info', None) is None:
          continue

        # Note that because pytsk3.TSK_FS_FILE does not explicitly define meta
        # we need to check if the attribute exists and has a value other
        # than None.
        if getattr(tsk_directory_entry.info, 'meta', None) is None:
          # Most directory entries will have an "inode" but not all, e.g.
          # previously deleted files. Currently directory entries without
          # a pytsk3.TSK_FS_META object are ignored.
          continue

        # Note that because pytsk3.TSK_FS_META does not explicitly define addr
        # we need to check if the attribute exists.
        if not hasattr(tsk_directory_entry.info.meta, 'addr'):
          continue

        directory_entry_inode = tsk_directory_entry.info.meta.addr
        directory_entry = None

        # Ignore references to self.
        if directory_entry_inode == inode:
          continue

        # On non-NTFS file systems ignore inode 0.
        if directory_entry_inode == 0 and not self._file_system.IsNTFS():
          continue

        # Note that because pytsk3.TSK_FS_FILE does not explicitly define name
        # we need to check if the attribute exists and has a value other
        # than None.
        if getattr(tsk_directory_entry.info, 'name', None) is not None:
          # Ignore file entries marked as "unallocated".
          flags = getattr(tsk_directory_entry.info.name, 'flags', 0)
          if int(flags) & pytsk3.TSK_FS_NAME_FLAG_UNALLOC:
            continue

          directory_entry = getattr(tsk_directory_entry.info.name, 'name', '')

          try:
            # pytsk3 returns an UTF-8 encoded byte string.
            directory_entry = directory_entry.decode('utf8')
          except UnicodeError:
            # Continue here since we cannot represent the directory entry.
            continue

          if directory_entry:
            # Ignore references to self or parent.
            if directory_entry in ['.', '..']:
              continue

            if not location or location == self._file_system.PATH_SEPARATOR:
              directory_entry = self._file_system.JoinPath([directory_entry])
            else:
              directory_entry = self._file_system.JoinPath([
                  location, directory_entry])

        yield tsk_path_spec.TSKPathSpec(
            inode=directory_entry_inode, location=directory_entry,
            parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""The HFS directory implementation."""

from dfvfs.path import hfs_path_spec
from dfvfs.vfs import directory


class HFSDirectory(directory.Directory):
  """File system directory that uses pyfshfs."""

  def __init__(self, file_system, path_spec, fshfs_file_entry):
    """Initializes a directory.

    Args:
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fshfs_file_entry (pyfshfs.file_entry): HFS file entry.
    """
    super(HFSDirectory, self).__init__(file_system, path_spec)
    self._fshfs_file_entry = fshfs_file_entry

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      HFSPathSpec: HFS path specification.
    """
    location = getattr(self.path_spec, 'location', None)

    for fshfs_sub_file_entry in self._fshfs_file_entry.sub_file_entries:
      directory_entry = fshfs_sub_file_entry.name

      if not location or location == self._file_system.PATH_SEPARATOR:
        directory_entry = self._file_system.JoinPath([directory_entry])
      else:
        directory_entry = self._file_system.JoinPath([
            location, directory_entry])

      yield hfs_path_spec.HFSPathSpec(
          identifier=fshfs_sub_file_entry.identifier, location=directory_entry,
          parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""The LUKSDE file entry implementation."""

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.lib import luksde_helper
from dfvfs.resolver import resolver
from dfvfs.vfs import root_only_file_entry


class LUKSDEFileEntry(root_only_file_entry.RootOnlyFileEntry):
  """File system file entry that uses pyluksde."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_LUKSDE

  def __init__(
      self, resolver_context, file_system, path_spec, is_root=False,
      is_virtual=False):
    """Initializes the file entry object.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file

    Raises:
      BackEndError: when the LUKSDE volume is missing.
    """
    luksde_volume = file_system.GetLUKSDEVolume()
    if luksde_volume is None:
      raise errors.BackEndError('Missing LUKSDE volume.')

    super(LUKSDEFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._luksde_volume = luksde_volume
    self.entry_type = definitions.FILE_ENTRY_TYPE_FILE

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    return self._luksde_volume.get_size()

  def IsLocked(self):
    """Determines if the file entry is locked.

    Returns:
      bool: True if the file entry is locked.
    """
    return self._luksde_volume.is_locked()

  def Unlock(self):
    """Unlocks the file entry.

    Returns:
      bool: True if the file entry was unlocked.
    """
    return luksde_helper.LUKSDEUnlockVolume(
        self._luksde_volume, self.path_spec, resolver.Resolver.key_chain)



# -*- coding: utf-8 -*-
"""The fake file system implementation."""

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import fake_path_spec
from dfvfs.vfs import file_system
from dfvfs.vfs import fake_file_entry


class FakeFileSystem(file_system.FileSystem):
  """Fake file system."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_FAKE

  def __init__(self, resolver_context, path_spec):
    """Initializes a file system.

    Args:
      resolver_context (Context): a resolver context.
      path_spec (PathSpec): a path specification.
    """
    super(FakeFileSystem, self).__init__(resolver_context, path_spec)
    self._paths = {}

    self.AddFileEntry(
        self.LOCATION_ROOT,
        file_entry_type=definitions.FILE_ENTRY_TYPE_DIRECTORY)

  def _Close(self):
    """Closes the file system.

    Raises:
      IOError: if the close failed.
    """
    return

  def _Open(self, mode='rb'):
    """Opens the file system defined by path specification.

    Args:
      mode (Optional[str]): file access mode. The default is 'rb' which
          represents read-only binary.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file system could not be opened.
      PathSpecError: if the path specification is incorrect.
      ValueError: if the path specification is invalid.
    """
    if self._path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification with parent.')

  def AddFileEntry(
      self, path, file_entry_type=definitions.FILE_ENTRY_TYPE_FILE,
      file_data=None, link_data=None):
    """Adds a fake file entry.

    Args:
      path (str): path of the file entry.
      file_entry_type (Optional[str]): type of the file entry object.
      file_data (Optional[bytes]): data of the fake file-like object.
      link_data (Optional[bytes]): link data of the fake file entry object.

    Raises:
      KeyError: if the path already exists.
      ValueError: if the file data is set but the file entry type is not a file
          or if the link data is set but the file entry type is not a link.
    """
    if path in self._paths:
      raise KeyError(f'File entry already set for path: {path:s}.')

    if file_data and file_entry_type != definitions.FILE_ENTRY_TYPE_FILE:
      raise ValueError('File data set for non-file file entry type.')

    if link_data and file_entry_type != definitions.FILE_ENTRY_TYPE_LINK:
      raise ValueError('Link data set for non-link file entry type.')

    if file_data is not None:
      path_data = file_data
    elif link_data is not None:
      path_data = link_data
    else:
      path_data = None

    self._paths[path] = (file_entry_type, path_data)

  def FileEntryExistsByPath(self, path):
    """Determines if a file entry for a path exists.

    Args:
      path (str): path of the file entry.

    Returns:
      bool: True if the file entry exists.
    """
    return path and path in self._paths

  def FileEntryExistsByPathSpec(self, path_spec):
    """Determines if a file entry for a path specification exists.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      bool: True if the file entry exists.
    """
    location = getattr(path_spec, 'location', None)
    return self.FileEntryExistsByPath(location)

  def GetDataByPath(self, path):
    """Retrieves the data associated to a path.

    Args:
      path (str): path of the file entry.

    Returns:
      bytes: data or None if not available.
    """
    _, path_data = self._paths.get(path, (None, None))
    return path_data

  def GetFileEntryByPath(self, path):
    """Retrieves a file entry for a path.

    Args:
      path (str): path of the file entry.

    Returns:
      FakeFileEntry: a file entry or None if not available.
    """
    if path is None:
      return None

    file_entry_type, _ = self._paths.get(path, (None, None))
    if not file_entry_type:
      return None

    path_spec = fake_path_spec.FakePathSpec(location=path)
    is_root = bool(path == self.LOCATION_ROOT)
    return fake_file_entry.FakeFileEntry(
        self._resolver_context, self, path_spec,
        file_entry_type=file_entry_type, is_root=is_root)

  def GetFileEntryByPathSpec(self, path_spec):
    """Retrieves a file entry for a path specification.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      FakeFileEntry: a file entry or None if not available.
    """
    location = getattr(path_spec, 'location', None)
    return self.GetFileEntryByPath(location)

  def GetPaths(self):
    """Retrieves the paths dictionary.

    Returns:
      dict[str, FileIO]: file-like object per path.
    """
    return self._paths

  def GetRootFileEntry(self):
    """Retrieves the root file entry.

    Returns:
      FakeFileEntry: a file entry or None if not available.
    """
    path_spec = fake_path_spec.FakePathSpec(location=self.LOCATION_ROOT)
    return self.GetFileEntryByPathSpec(path_spec)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Volume Shadow Snapshots (VSS) volume system."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.volume import vshadow_volume_system

from tests import test_lib as shared_test_lib


class VShadowVolumeSystemTest(shared_test_lib.BaseTestCase):
  """Tests the Volume Shadow Snapshot (VSS) volume system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['vss.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._vshadow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_VSHADOW, location='/',
        parent=test_raw_path_spec)

  # vshadowinfo test_data/vss.raw
  # vshadowinfo 20210425
  #
  # Volume Shadow Snapshot information:
  #     Number of stores:	2
  #
  # Store: 1
  #     Identifier          : de81cc22-aa8b-11eb-9339-8cdcd4557abc
  #     Shadow copy set ID  : 6c5c9cd2-ea46-4c70-a4a8-568fdabd27c1
  #     Creation time       : May 01, 2021 17:40:03.223030400 UTC
  #     Shadow copy ID      : 2c6c6cc8-2b97-41da-a030-4add838ae8f6
  #     Volume size         : 78 MiB (82771968 bytes)
  #     Attribute flags     : 0x00420009
  #
  # Store: 2
  #     Identifier          : de81cc2b-aa8b-11eb-9339-8cdcd4557abc
  #     Shadow copy set ID  : b4f4b9d6-1cf2-4bfc-b1a3-c2f6e9628ef9
  #     Creation time       : May 01, 2021 17:41:28.224986300 UTC
  #     Shadow copy ID      : 19e1881a-c184-4ec4-908e-766ba3373e8a
  #     Volume size         : 78 MiB (82771968 bytes)
  #     Attribute flags     : 0x00420009

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = vshadow_volume_system.VShadowVolumeSystem()

    volume_system.Open(self._vshadow_path_spec)

    self.assertEqual(volume_system.number_of_sections, 0)
    self.assertEqual(volume_system.number_of_volumes, 2)

    self.assertEqual(volume_system.volume_identifiers, ['vss1', 'vss2'])

    volume = volume_system.GetVolumeByIndex(1)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 4)
    self.assertEqual(volume.identifier, 'vss2')

    expected_value = 'de81cc2b-aa8b-11eb-9339-8cdcd4557abc'
    volume_attribute = volume.GetAttribute('identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    expected_value = '19e1881a-c184-4ec4-908e-766ba3373e8a'
    volume_attribute = volume.GetAttribute('copy_identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    expected_value = 'b4f4b9d6-1cf2-4bfc-b1a3-c2f6e9628ef9'
    volume_attribute = volume.GetAttribute('copy_set_identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    expected_value = 132643644882249863
    volume_attribute = volume.GetAttribute('creation_time')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    volume = volume_system.GetVolumeByIndex(7)
    self.assertIsNone(volume)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Virtual File System (VFS) path specification interface."""

import unittest

from dfvfs.path import path_spec

from tests.path import test_lib


class TestPathSpec(path_spec.PathSpec):
  """Path specification for testing."""

  TYPE_INDICATOR = 'test'

  def __init__(self, parent=None, **kwargs):
    """Initializes a path specification.

    Args:
      parent (Optional[PathSpec]): parent path specification.
    """
    super(TestPathSpec, self).__init__(parent=parent, **kwargs)
    self.attribute = 'MyAttribute'


class PathSpecTest(test_lib.PathSpecTestCase):
  """Tests for the VFS path specification interface."""

  # pylint: disable=protected-access

  def testInitialize(self):
    """Tests the __init__ function."""
    with self.assertRaises(ValueError):
      path_spec.PathSpec()

  # TODO: add tests for __eq__
  # TODO: add tests for __hash__

  def testGetComparable(self):
    """Tests the _GetComparable function."""
    test_path_spec = TestPathSpec()

    test_comparable = test_path_spec._GetComparable()
    self.assertEqual(test_comparable, 'type: test\n')

  def testComparable(self):
    """Tests the comparable property."""
    test_path_spec = TestPathSpec()

    self.assertEqual(test_path_spec.comparable, 'type: test\n')

  def testTypeIndicator(self):
    """Tests the type_indicator property."""
    test_path_spec = TestPathSpec()

    self.assertEqual(test_path_spec.type_indicator, 'test')

  def testCopyToDict(self):
    """Tests the CopyToDict function."""
    test_path_spec = TestPathSpec()

    test_dict = test_path_spec.CopyToDict()
    self.assertEqual(test_dict, {'attribute': 'MyAttribute'})

  def testHasParent(self):
    """Tests the HasParent function."""
    test_path_spec = TestPathSpec()

    self.assertFalse(test_path_spec.HasParent())

  def testIsFileSystem(self):
    """Tests the IsFileSystem function."""
    test_path_spec = TestPathSpec()

    self.assertFalse(test_path_spec.IsFileSystem())

  def testIsSystemLevel(self):
    """Tests the IsSystemLevel function."""
    test_path_spec = TestPathSpec()

    self.assertFalse(test_path_spec.IsSystemLevel())

  def testIsVolumeSystem(self):
    """Tests the IsVolumeSystem function."""
    test_path_spec = TestPathSpec()

    self.assertFalse(test_path_spec.IsVolumeSystem())

  def testIsVolumeSystemRoot(self):
    """Tests the IsVolumeSystemRoot function."""
    test_path_spec = TestPathSpec()

    self.assertFalse(test_path_spec.IsVolumeSystemRoot())


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The fake directory implementation."""

from dfvfs.path import fake_path_spec
from dfvfs.vfs import directory


class FakeDirectory(directory.Directory):
  """Fake file system directory."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      FakePathSpec: a path specification.
    """
    location = getattr(self.path_spec, 'location', None)
    if location is not None:
      paths = self._file_system.GetPaths()

      for path in paths.keys():
        # Determine if the start of the path is similar to the location string.
        # If not the file the path refers to is not in the same directory.
        if not path or not path.startswith(location):
          continue

        _, suffix = self._file_system.GetPathSegmentAndSuffix(location, path)

        # Ignore anything that is part of a sub directory or the directory
        # itself.
        if suffix or path == location:
          continue

        path_spec_location = self._file_system.JoinPath([path])
        yield fake_path_spec.FakePathSpec(location=path_spec_location)



# -*- coding: utf-8 -*-
"""The Virtual File System (VFS) data stream interface."""


class DataStream(object):
  """Data stream interface."""

  def __init__(self, file_entry):
    """Initializes the data stream.

    Args:
      file_entry (FileEntry): file entry.
    """
    super(DataStream, self).__init__()
    self._file_entry = file_entry
    self._name = ''

  @property
  def name(self):
    """str: name."""
    return self._name

  def GetExtents(self):
    """Retrieves the extents.

    Returns:
      list[Extent]: the extents of the data stream.
    """
    if self._file_entry:
      return self._file_entry.GetExtents()
    return []

  def IsDefault(self):
    """Determines if the data stream is the default data stream.

    Returns:
      bool: True if the data stream is the default data stream.
    """
    return not bool(self._name)



# -*- coding: utf-8 -*-
"""The operating system file-like object implementation."""

import stat
import os

import pysmdev

from dfvfs.file_io import file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors


class OSFile(file_io.FileIO):
  """File input/output (IO) object that uses the operating system."""

  def __init__(self, resolver_context, path_spec):
    """Initializes a file input/output (IO) object.

    Args:
      resolver_context (Context): resolver context.
      path_spec (PathSpec): a path specification.
    """
    super(OSFile, self).__init__(resolver_context, path_spec)
    self._file_object = None
    self._size = 0

  def _Close(self):
    """Closes the file-like object."""
    self._file_object.close()
    self._file_object = None

  def _Open(self, mode='rb'):
    """Opens the file-like object defined by path specification.

    Args:
      mode (Optional[str]): file access mode.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file-like object could not be opened.
      MountPointError: if the mount point specified in the path specification
          does not exist.
      OSError: if the file-like object could not be opened.
      PathSpecError: if the path specification is incorrect.
    """
    location = getattr(self._path_spec, 'location', None)

    if self._path_spec.HasParent():
      parent_path_spec = self._path_spec.parent
      if parent_path_spec.type_indicator != definitions.TYPE_INDICATOR_MOUNT:
        raise errors.PathSpecError(
            'Unsupported path specification with parent.')

      mount_path_spec = self._resolver_context.GetMountPoint(
          parent_path_spec.identifier)
      if not mount_path_spec:
        raise errors.MountPointError(
            f'No such mount point: {parent_path_spec.identifier:s}')

      if (mount_path_spec.type_indicator != definitions.TYPE_INDICATOR_OS or
          mount_path_spec.parent):
        raise errors.MountPointError(
            'Unsupported mount point path specification.')

      mount_path = mount_path_spec.location
      if mount_path[-1] == os.path.sep:
        mount_path = mount_path[:-1]

      # Cannot use os.path.join() here since location is prefixed with a path
      # segment separator.
      location = ''.join([mount_path, location])

    if location is None:
      raise errors.PathSpecError('Path specification missing location.')

    # Windows does not support running os.stat on device files so we use
    # libsmdev to do an initial check.
    try:
      is_device = pysmdev.check_device(location)
    except IOError as exception:
      # Since os.stat() will not recognize Windows device file names and
      # will return '[Error 87] The parameter is incorrect' we check here
      # if pysmdev exception message contains ' access denied ' and raise
      # AccessError instead.
      if ' access denied ' in str(exception):
        raise errors.AccessError(
            f'Access denied to file: {location:s} with error: {exception!s}')

      is_device = False

    if not is_device:
      try:
        stat_info = os.stat(location)
      except OSError as exception:
        raise IOError(f'Unable to open file with error: {exception!s}')

      # In case the libsmdev check is not able to detect the device also use
      # the stat information.
      if stat.S_ISCHR(stat_info.st_mode) or stat.S_ISBLK(stat_info.st_mode):
        is_device = True

    if is_device:
      smdev_handle = pysmdev.handle()
      smdev_handle.open(location, mode=mode)

      self._file_object = smdev_handle
      self._size = smdev_handle.media_size

    else:
      self._file_object = open(location, mode=mode)  # pylint: disable=consider-using-with,unspecified-encoding
      self._size = stat_info.st_size

  # Note: that the following functions do not follow the style guide
  # because they are part of the file-like object interface.
  # pylint: disable=invalid-name

  def read(self, size=None):
    """Reads a byte string from the file-like object at the current offset.

    The function will read a byte string of the specified size or
    all of the remaining data if no size was specified.

    Args:
      size (Optional[int]): number of bytes to read, where None is all
          remaining data.

    Returns:
      bytes: data read.

    Raises:
      IOError: if the read failed.
      OSError: if the read failed.
    """
    if not self._is_open:
      raise IOError('Not opened.')

    if size is None:
      size = self._size - self._file_object.tell()

    return self._file_object.read(size)

  def seek(self, offset, whence=os.SEEK_SET):
    """Seeks to an offset within the file-like object.

    Args:
      offset (int): offset to seek to.
      whence (Optional(int)): value that indicates whether offset is an absolute
          or relative position within the file.

    Raises:
      IOError: if the seek failed.
      OSError: if the seek failed.
    """
    if not self._is_open:
      raise IOError('Not opened.')

    # For a yet unknown reason a Python file-like object on Windows allows for
    # invalid whence values to be passed to the seek function. This check
    # makes sure the behavior of the function is the same on all platforms.
    if whence not in (os.SEEK_CUR, os.SEEK_END, os.SEEK_SET):
      raise IOError('Unsupported whence.')

    self._file_object.seek(offset, whence)

  def get_offset(self):
    """Retrieves the current offset into the file-like object.

    Returns:
      int: current offset into the file-like object.

    Raises:
      IOError: if the file-like object has not been opened.
      OSError: if the file-like object has not been opened.
    """
    if not self._is_open:
      raise IOError('Not opened.')

    return self._file_object.tell()

  def get_size(self):
    """Retrieves the size of the file-like object.

    Returns:
      int: size of the file-like object data.

    Raises:
      IOError: if the file-like object has not been opened.
      OSError: if the file-like object has not been opened.
    """
    if not self._is_open:
      raise IOError('Not opened.')

    return self._size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the attribute implementation using pytsk3."""

import unittest

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import tsk_attribute
from dfvfs.vfs import tsk_file_system

from tests import test_lib as shared_test_lib


class TSKAttributeTest(shared_test_lib.BaseTestCase):
  """Tests the TSK attribute."""

  # pylint: disable=protected-access

  _INODE_A_FILE = 19

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['hfsplus.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/',
        parent=self._raw_path_spec)

    self._file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testIntialize(self):
    """Tests the __init__ function."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._INODE_A_FILE,
        location='/a_directory/a_file', parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
    self.assertIsNotNone(file_entry)

    pytsk_attribute = next(file_entry._tsk_file)
    self.assertIsNotNone(pytsk_attribute)

    test_attribute = tsk_attribute.TSKAttribute(
        file_entry._tsk_file, pytsk_attribute)
    self.assertIsNotNone(test_attribute)

    with self.assertRaises(errors.BackEndError):
      tsk_attribute.TSKAttribute(None, pytsk_attribute)

    with self.assertRaises(errors.BackEndError):
      tsk_attribute.TSKAttribute(file_entry._tsk_file, None)


class TSKExtendedAttributeTest(shared_test_lib.BaseTestCase):
  """Tests the TSK extended attribute."""

  # pylint: disable=protected-access

  _INODE_A_FILE = 19

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['hfsplus.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/',
        parent=self._raw_path_spec)

    self._file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testIntialize(self):
    """Tests the __init__ function."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._INODE_A_FILE,
        location='/a_directory/a_file', parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
    self.assertIsNotNone(file_entry)

    pytsk_attribute = next(file_entry._tsk_file)
    self.assertIsNotNone(pytsk_attribute)
    self.assertEqual(pytsk_attribute.info.name, b'myxattr')

    test_attribute = tsk_attribute.TSKExtendedAttribute(
        file_entry._tsk_file, pytsk_attribute)
    self.assertIsNotNone(test_attribute)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Helper function for Core Storage (CS) support."""


_CS_LOCATION_PREFIX = '/cs'
_CS_LOCATION_PREFIX_LENGTH = len(_CS_LOCATION_PREFIX)


def CSPathSpecGetVolumeIndex(path_spec):
  """Retrieves the volume index from the path specification.

  Args:
    path_spec (PathSpec): path specification.

  Returns:
    int: volume index or None if not available.
  """
  volume_index = getattr(path_spec, 'volume_index', None)

  if volume_index is None:
    location = getattr(path_spec, 'location', None)

    if location is None or not location.startswith(_CS_LOCATION_PREFIX):
      return None

    volume_index = None
    try:
      volume_index = int(location[_CS_LOCATION_PREFIX_LENGTH:], 10) - 1
    except ValueError:
      pass

    if volume_index is None or volume_index < 0:
      return None

  return volume_index


def CSUnlockLogicalVolume(fvde_logical_volume, path_spec, key_chain):
  """Unlocks a Core Storage logical volume using the path specification.

  Args:
    fvde_logical_volume (pyfvde.logical_volume): Core Storage logical volume.
    path_spec (PathSpec): path specification.
    key_chain (KeyChain): key chain.

  Returns:
    bool: True if the volume is unlocked, False otherwise.
  """
  is_locked = fvde_logical_volume.is_locked()
  if is_locked:
    password = key_chain.GetCredential(path_spec, 'password')
    if password:
      fvde_logical_volume.set_password(password)

    recovery_password = key_chain.GetCredential(path_spec, 'recovery_password')
    if recovery_password:
      fvde_logical_volume.set_recovery_password(recovery_password)

    is_locked = not fvde_logical_volume.unlock()

  return not is_locked



# -*- coding: utf-8 -*-
"""The fake path specification resolver helper implementation."""

from dfvfs.lib import definitions
from dfvfs.resolver_helpers import manager
from dfvfs.resolver_helpers import resolver_helper


class FakeResolverHelper(resolver_helper.ResolverHelper):
  """Fake resolver helper."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_FAKE

  # The fake file system resolver has no implementation and
  # should raise RuntimeError.


manager.ResolverHelperManager.RegisterHelper(FakeResolverHelper())



# -*- coding: utf-8 -*-
"""The triple DES decrypter implementation."""

import pyfcrypto

from dfvfs.encryption import decrypter
from dfvfs.encryption import manager
from dfvfs.lib import definitions


class DES3Decrypter(decrypter.Decrypter):
  """Triple DES decrypter using pyfcrypto."""

  ENCRYPTION_METHOD = definitions.ENCRYPTION_METHOD_DES3

  _BLOCK_SIZE = 8

  # TODO: add CFB and OFB support

  _ENCRYPTION_MODES = frozenset([
      definitions.ENCRYPTION_MODE_CBC,
      definitions.ENCRYPTION_MODE_ECB])

  def __init__(
      self, cipher_mode=None, initialization_vector=None, key=None, **kwargs):
    """Initializes a decrypter.

    Args:
      cipher_mode (Optional[str]): cipher mode.
      initialization_vector (Optional[bytes]): initialization vector.
      key (Optional[bytes]): key.
      kwargs (dict): keyword arguments depending on the decrypter.

    Raises:
      ValueError: when key is not set, the cipher mode is not supported or
          initialization vector not set or not supported.
      TypeError: when the initialization vector type is not supported.
    """
    if not key:
      raise ValueError('Missing key.')

    if cipher_mode not in self._ENCRYPTION_MODES:
      raise ValueError(f'Unsupported cipher mode: {cipher_mode!s}')

    if (cipher_mode != definitions.ENCRYPTION_MODE_ECB and
        not initialization_vector):
      raise ValueError('Missing initialization vector.')

    if initialization_vector and not isinstance(initialization_vector, bytes):
      raise TypeError('Unsupported initialization vector type.')

    if initialization_vector and len(initialization_vector) != self._BLOCK_SIZE:
      raise ValueError('Unsupported initialization vector size.')

    super(DES3Decrypter, self).__init__()
    self._des3_context = pyfcrypto.des3_context()
    self._cipher_mode = cipher_mode
    self._initialization_vector = initialization_vector

    self._des3_context.set_key(key)

  def Decrypt(self, encrypted_data, finalize=False):
    """Decrypts the encrypted data.

    Args:
      encrypted_data (bytes): encrypted data.
      finalize (Optional[bool]): True if the end of data has been reached and
          the cipher context should be finalized.

    Returns:
      tuple[bytes, bytes]: decrypted data and remaining encrypted data.
    """
    encrypted_data_size = len(encrypted_data)

    _, block_offset = divmod(encrypted_data_size, self._BLOCK_SIZE)

    remaining_encrypted_data = b''
    if block_offset > 0:
      if finalize:
        block_padding_size = self._BLOCK_SIZE - block_offset
        encrypted_data = b''.join([
            encrypted_data, b'\x00' * block_padding_size])
      else:
        remaining_encrypted_data = encrypted_data[-block_offset:]
        encrypted_data = encrypted_data[:-block_offset]

    if self._cipher_mode == definitions.ENCRYPTION_MODE_CBC:
      decrypted_data = pyfcrypto.crypt_des3_cbc(
          self._des3_context, pyfcrypto.crypt_modes.DECRYPT,
          self._initialization_vector, encrypted_data)

      self._initialization_vector = encrypted_data[-self._BLOCK_SIZE:]

    elif self._cipher_mode == definitions.ENCRYPTION_MODE_ECB:
      decrypted_data = pyfcrypto.crypt_des3_ecb(
          self._des3_context, pyfcrypto.crypt_modes.DECRYPT,
          encrypted_data)

    decrypted_data_size = encrypted_data_size - block_offset
    return decrypted_data[:decrypted_data_size], remaining_encrypted_data


manager.EncryptionManager.RegisterDecrypter(DES3Decrypter)



# -*- coding: utf-8 -*-
"""The BitLocker Drive Encryption (BDE) credentials."""

from dfvfs.credentials import credentials
from dfvfs.credentials import manager
from dfvfs.lib import definitions


# TODO: add means to set the credentials in the bde_volume using the helper?


class BDECredentials(credentials.Credentials):
  """BitLocker Drive Encryption (BDE) credentials."""

  # TODO: add support for key_data credential, needs pybde update.
  CREDENTIALS = frozenset([
      'password', 'recovery_password', 'startup_key'])

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_BDE


manager.CredentialsManager.RegisterCredentials(BDECredentials())



# -*- coding: utf-8 -*-
"""The NTFS file entry implementation."""

import copy

import pyfsntfs
import pyfwnt

from dfdatetime import filetime as dfdatetime_filetime

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import ntfs_path_spec
from dfvfs.resolver import resolver
from dfvfs.vfs import attribute
from dfvfs.vfs import extent
from dfvfs.vfs import file_entry
from dfvfs.vfs import ntfs_attribute
from dfvfs.vfs import ntfs_data_stream
from dfvfs.vfs import ntfs_directory


class NTFSFileEntry(file_entry.FileEntry):
  """File system file entry that uses pyfsntfs."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_NTFS

  _ATTRIBUTE_TYPE_CLASS_MAPPINGS = {
      0x00000010: ntfs_attribute.StandardInformationNTFSAttribute,
      0x00000030: ntfs_attribute.FileNameNTFSAttribute,
      0x00000040: ntfs_attribute.ObjectIdentifierNTFSAttribute,
      0x00000050: ntfs_attribute.SecurityDescriptorNTFSAttribute,
  }

  _FILE_REFERENCE_MFT_ENTRY_BITMASK = 0xffffffffffff

  def __init__(
      self, resolver_context, file_system, path_spec, fsntfs_file_entry=None,
      is_root=False, is_virtual=False):
    """Initializes the file entry object.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsntfs_file_entry (Optional[pyfsntfs.file_entry]): NTFS file entry.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file
          entry emulated by the corresponding file system.

    Raises:
      BackEndError: if the pyfsntfs file entry is missing.
    """
    if not fsntfs_file_entry:
      fsntfs_file_entry = file_system.GetNTFSFileEntryByPathSpec(path_spec)
    if not fsntfs_file_entry:
      raise errors.BackEndError('Missing pyfsntfs file entry.')

    super(NTFSFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._fsntfs_file_entry = fsntfs_file_entry

    file_attribute_flags = fsntfs_file_entry.file_attribute_flags
    if self._IsLink(file_attribute_flags):
      self.entry_type = definitions.FILE_ENTRY_TYPE_LINK
    elif fsntfs_file_entry.has_directory_entries_index():
      self.entry_type = definitions.FILE_ENTRY_TYPE_DIRECTORY
    elif self._IsDevice(file_attribute_flags):
      self.entry_type = definitions.FILE_ENTRY_TYPE_DEVICE
    else:
      self.entry_type = definitions.FILE_ENTRY_TYPE_FILE

  def _GetAttributes(self):
    """Retrieves the attributes.

    Returns:
      list[Attribute]: attributes.
    """
    if self._attributes is None:
      self._attributes = []

      for fsntfs_attribute in self._fsntfs_file_entry.attributes:
        attribute_class = self._ATTRIBUTE_TYPE_CLASS_MAPPINGS.get(
            fsntfs_attribute.attribute_type, ntfs_attribute.NTFSAttribute)

        attribute_object = attribute_class(fsntfs_attribute)
        self._attributes.append(attribute_object)

    return self._attributes

  def _GetDataStreams(self):
    """Retrieves the data streams.

    Returns:
      list[NTFSDataStream]: data streams.
    """
    if self._data_streams is None:
      self._data_streams = []
      if self._fsntfs_file_entry.has_default_data_stream():
        data_stream = ntfs_data_stream.NTFSDataStream(self, None)
        self._data_streams.append(data_stream)

      for fsntfs_data_stream in self._fsntfs_file_entry.alternate_data_streams:
        data_stream = ntfs_data_stream.NTFSDataStream(self, fsntfs_data_stream)
        self._data_streams.append(data_stream)

    return self._data_streams

  def _GetDirectory(self):
    """Retrieves a directory.

    Returns:
      NTFSDirectory: a directory.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_DIRECTORY:
      return None

    return ntfs_directory.NTFSDirectory(
        self._file_system, self.path_spec, self._fsntfs_file_entry)

  def _GetLink(self):
    """Retrieves the link.

    Returns:
      str: path of the linked file.
    """
    if self._link is None:
      self._link = self._fsntfs_file_entry.symbolic_link_target
      if self._link:
        # Strip off the drive letter, we assume the link is within
        # the same volume.
        _, _, self._link = self._link.rpartition(':')

    return self._link

  def _GetStatAttribute(self):
    """Retrieves a stat attribute.

    Returns:
      StatAttribute: a stat attribute or None if not available.
    """
    stat_attribute = attribute.StatAttribute()
    stat_attribute.inode_number = getattr(
        self._fsntfs_file_entry, 'file_reference', None)
    stat_attribute.number_of_links = getattr(
        self._fsntfs_file_entry, 'number_of_links', None)
    stat_attribute.size = getattr(self._fsntfs_file_entry, 'size', None)
    stat_attribute.type = self.entry_type

    return stat_attribute

  def _GetSubFileEntries(self):
    """Retrieves sub file entries.

    Yields:
      NTFSFileEntry: a sub file entry.
    """
    if self._directory is None:
      self._directory = self._GetDirectory()

    if self._directory:
      for path_spec in self._directory.entries:
        yield NTFSFileEntry(
            self._resolver_context, self._file_system, path_spec)

  def _IsDevice(self, file_attribute_flags):
    """Determines if a file entry is a device.

    Args:
      file_attribute_flags (int): file attribute flags.

    Returns:
      bool: True if a file entry is device link, false otherwise.
    """
    if file_attribute_flags is None:
      return False
    return bool(file_attribute_flags & pyfsntfs.file_attribute_flags.DEVICE)

  def _IsLink(self, file_attribute_flags):
    """Determines if a file entry is a link.

    Args:
      file_attribute_flags (int): file attribute flags.

    Returns:
      bool: True if a file entry is a link, false otherwise.
    """
    if file_attribute_flags is None:
      return False
    return bool(
        file_attribute_flags & pyfsntfs.file_attribute_flags.REPARSE_POINT)

  @property
  def access_time(self):
    """dfdatetime.DateTimeValues: access time or None if not available."""
    timestamp = self._fsntfs_file_entry.get_access_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def change_time(self):
    """dfdatetime.DateTimeValues: change time or None if not available."""
    timestamp = self._fsntfs_file_entry.get_entry_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def creation_time(self):
    """dfdatetime.DateTimeValues: creation time or None if not available."""
    timestamp = self._fsntfs_file_entry.get_creation_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def modification_time(self):
    """dfdatetime.DateTimeValues: modification time or None if not available."""
    timestamp = self._fsntfs_file_entry.get_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def name(self):
    """str: name of the file entry, which does not include the full path."""
    # The root directory file name is typically '.', dfVFS however uses ''.
    if self._is_root:
      return ''

    mft_attribute = getattr(self.path_spec, 'mft_attribute', None)
    if mft_attribute is not None:
      return self._fsntfs_file_entry.get_name_by_attribute_index(mft_attribute)

    return self._fsntfs_file_entry.get_name()

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    return getattr(self._fsntfs_file_entry, 'size', None)

  def GetExtents(self):
    """Retrieves the extents.

    Returns:
      list[Extent]: the extents.
    """
    extents = []
    for extent_index in range(self._fsntfs_file_entry.number_of_extents):
      extent_offset, extent_size, extent_flags = (
          self._fsntfs_file_entry.get_extent(extent_index))

      if extent_flags & 0x1:
        extent_type = definitions.EXTENT_TYPE_SPARSE
      else:
        extent_type = definitions.EXTENT_TYPE_DATA

      data_stream_extent = extent.Extent(
          extent_type=extent_type, offset=extent_offset, size=extent_size)
      extents.append(data_stream_extent)

    return extents

  def GetFileObject(self, data_stream_name=''):
    """Retrieves a file-like object of a specific data stream.

    Args:
      data_stream_name (Optional[str]): data stream name, where an empty
          string represents the default data stream.

    Returns:
      NTFSFileIO: file-like object or None.
    """
    if (not data_stream_name and
        not self._fsntfs_file_entry.has_default_data_stream()):
      return None

    # Make sure to make the changes on a copy of the path specification, so we
    # do not alter self.path_spec.
    path_spec = copy.deepcopy(self.path_spec)
    if data_stream_name:
      setattr(path_spec, 'data_stream', data_stream_name)

    return resolver.Resolver.OpenFileObject(
        path_spec, resolver_context=self._resolver_context)

  def GetLinkedFileEntry(self):
    """Retrieves the linked file entry, e.g. for a symbolic link.

    Returns:
      NTFSFileEntry: linked file entry or None.
    """
    link = self._GetLink()
    if not link:
      return None

    # TODO: is there a way to determine the MFT entry here?
    link_mft_entry = None

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    path_spec = ntfs_path_spec.NTFSPathSpec(
        location=link, parent=parent_path_spec)

    is_root = bool(
        link == self._file_system.LOCATION_ROOT or
        link_mft_entry == self._file_system.MFT_ENTRY_ROOT_DIRECTORY)

    return NTFSFileEntry(
        self._resolver_context, self._file_system, path_spec, is_root=is_root)

  def GetNTFSFileEntry(self):
    """Retrieves the NTFS file entry.

    Returns:
      pyfsntfs.file_entry: NTFS file entry.
    """
    return self._fsntfs_file_entry

  def GetParentFileEntry(self):
    """Retrieves the parent file entry.

    Returns:
      NTFSFileEntry: parent file entry or None if not available.
    """
    location = getattr(self.path_spec, 'location', None)
    if location is not None:
      parent_location = self._file_system.DirnamePath(location)
      if parent_location == '':
        parent_location = self._file_system.PATH_SEPARATOR

    parent_file_reference = None
    mft_attribute = getattr(self.path_spec, 'mft_attribute', None)
    if mft_attribute is not None:
      parent_file_reference = (
          self._fsntfs_file_entry.get_parent_file_reference_by_attribute_index(
              mft_attribute))
    else:
      parent_file_reference = (
          self._fsntfs_file_entry.get_parent_file_reference())

    if parent_file_reference is None:
      return None

    parent_mft_entry = (
        parent_file_reference & self._FILE_REFERENCE_MFT_ENTRY_BITMASK)

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    # TODO: determine and pass the mft_attribute of the parent
    # for a faster resolve of the file entry.
    path_spec = ntfs_path_spec.NTFSPathSpec(
        location=parent_location, mft_entry=parent_mft_entry,
        parent=parent_path_spec)

    # TODO: handle parent correctly use attribute index?
    is_root = bool(
        parent_location == self._file_system.LOCATION_ROOT or
        parent_mft_entry == self._file_system.MFT_ENTRY_ROOT_DIRECTORY)

    return NTFSFileEntry(
        self._resolver_context, self._file_system, path_spec, is_root=is_root)

  def GetSecurityDescriptor(self):
    """Retrieves the security descriptor.

    Returns:
      pyfwnt.security_descriptor: security descriptor.
    """
    fwnt_security_descriptor = pyfwnt.security_descriptor()
    fwnt_security_descriptor.copy_from_byte_stream(
        self._fsntfs_file_entry.security_descriptor_data)

    return fwnt_security_descriptor

  def IsAllocated(self):
    """Determines if the file entry is allocated.

    Returns:
      bool: True if the file entry is allocated.
    """
    return self._fsntfs_file_entry.is_allocated()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the encrypted stream resolver helper implementation."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver_helpers import encrypted_stream_resolver_helper

from tests.resolver_helpers import test_lib


class EncryptedStreamResolverHelperTest(test_lib.ResolverHelperTestCase):
  """Tests for the encrypted stream resolver helper implementation."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(EncryptedStreamResolverHelperTest, self).setUp()

    test_path = self._GetTestFilePath(['syslog.rc4'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._encrypted_stream_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_ENCRYPTED_STREAM,
        encryption_method=definitions.ENCRYPTION_METHOD_RC4,
        parent=test_os_path_spec)

  def testNewFileObject(self):
    """Tests the NewFileObject function."""
    resolver_helper_object = (
        encrypted_stream_resolver_helper.EncryptedStreamResolverHelper())
    self._TestNewFileObject(
        resolver_helper_object, self._encrypted_stream_path_spec)

  def testNewFileSystem(self):
    """Tests the NewFileSystem function."""
    resolver_helper_object = (
        encrypted_stream_resolver_helper.EncryptedStreamResolverHelper())
    self._TestNewFileSystem(
        resolver_helper_object, self._encrypted_stream_path_spec)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the compressed stream file system implementation."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import compressed_stream_file_system

from tests import test_lib as shared_test_lib


class CompressedStreamFileSystemTest(shared_test_lib.BaseTestCase):
  """Tests the compressed stream file system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.bz2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._compressed_stream_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_COMPRESSED_STREAM,
        compression_method=definitions.COMPRESSION_METHOD_BZIP2,
        parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenAndClose(self):
    """Test the open and close functionality."""
    file_system = compressed_stream_file_system.CompressedStreamFileSystem(
        self._resolver_context, self._compressed_stream_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

  def testFileEntryExistsByPathSpec(self):
    """Test the file entry exists by path specification functionality."""
    file_system = compressed_stream_file_system.CompressedStreamFileSystem(
        self._resolver_context, self._compressed_stream_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    self.assertTrue(file_system.FileEntryExistsByPathSpec(
        self._compressed_stream_path_spec))

  def testGetFileEntryByPathSpec(self):
    """Tests the GetFileEntryByPathSpec function."""
    file_system = compressed_stream_file_system.CompressedStreamFileSystem(
        self._resolver_context, self._compressed_stream_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    file_entry = file_system.GetFileEntryByPathSpec(
        self._compressed_stream_path_spec)

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, '')

  def testGetRootFileEntry(self):
    """Test the get root file entry functionality."""
    file_system = compressed_stream_file_system.CompressedStreamFileSystem(
        self._resolver_context, self._compressed_stream_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    file_entry = file_system.GetRootFileEntry()

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, '')


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the file system implementation using the SleuthKit (TSK)."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import tsk_file_system

from tests import test_lib as shared_test_lib


class TSKFileSystemTest(shared_test_lib.BaseTestCase):
  """Tests the SleuthKit (TSK) file system."""

  _INODE_PASSWORDS_TXT = 14

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['ext2.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/',
        parent=self._raw_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenAndClose(self):
    """Test the open and close functionality."""
    file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

  def testFileEntryExistsByPathSpec(self):
    """Test the file entry exists by path specification functionality."""
    file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._INODE_PASSWORDS_TXT,
        location='/password.txt', parent=self._raw_path_spec)
    self.assertTrue(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=9999, location='/bogus.txt',
        parent=self._raw_path_spec)
    self.assertFalse(file_system.FileEntryExistsByPathSpec(path_spec))

  def testGetFileEntryByPathSpec(self):
    """Tests the GetFileEntryByPathSpec function."""
    file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=15, parent=self._raw_path_spec)

    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNotNone(file_entry)
    # There is no way to determine the file_entry.name without a location string
    # in the path_spec or retrieving the file_entry from its parent.

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._INODE_PASSWORDS_TXT,
        location='/password.txt', parent=self._raw_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, 'password.txt')

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=9999, location='/bogus.txt',
        parent=self._raw_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNone(file_entry)

  def testGetRootFileEntry(self):
    """Test the get root file entry functionality."""
    file_system = tsk_file_system.TSKFileSystem(
        self._resolver_context, self._tsk_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    file_entry = file_system.GetRootFileEntry()

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, '')


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Logical Volume Manager (LVM) volume system."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.volume import lvm_volume_system

from tests import test_lib as shared_test_lib


class LVMVolumeSystemTest(shared_test_lib.BaseTestCase):
  """Tests the Logical Volume Manager (LVM) volume system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['lvm.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._lvm_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, location='/',
        parent=test_raw_path_spec)

  # vslvminfo lvm.raw
  # vslvminfo 20210524
  #
  # Linux Logical Volume Manager (LVM) information:
  # Volume Group (VG):
  #     Name:                          test_volume_group
  #     Identifier:                    Lu8jtE-dBED-jxdl-bAXG-vmAA-Oh9y-KCiNMy
  #     Sequence number:               3
  #     Extent size:                   4.0 MiB (4194304 bytes)
  #     Number of physical volumes:    1
  #     Number of logical volumes:     2
  #
  # Physical Volume (PV): 1
  #     Name:                          pv0
  #     Identifier:                    V9N6H5-MXlo-JmEE-TLHs-KyAo-kLv5-JL0RY7
  #     Device path:                   /dev/loop99
  #     Volume size:                   10 MiB (10485760 bytes)
  #
  # Logical Volume (LV): 1
  #     Name:                          test_logical_volume1
  #     Identifier:                    EMfcGR-KAOf-VrvI-yqth-fmz0-Ylmx-xepAJq
  #     Number of segments:            1
  #     Segment: 1
  #         Offset:                    0x00000000 (0)
  #         Size:                      4.0 MiB (4194304 bytes)
  #         Number of stripes:         1
  #         Stripe: 1
  #             Physical volume:       pv0
  #             Data area offset:      0x00000000 (0)
  #
  # Logical Volume (LV): 2
  #     Name:                          test_logical_volume2
  #     Identifier:                    UbmaM4-4OCn-m0mb-uw5D-kCjp-2hi1-M6kWCX
  #     Number of segments:            1
  #     Segment: 1
  #         Offset:                    0x00000000 (0)
  #         Size:                      4.0 MiB (4194304 bytes)
  #         Number of stripes:         1
  #         Stripe: 1
  #             Physical volume:       pv0
  #             Data area offset:      0x00400000 (4194304)

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = lvm_volume_system.LVMVolumeSystem()

    volume_system.Open(self._lvm_path_spec)

    self.assertEqual(volume_system.number_of_sections, 0)
    self.assertEqual(volume_system.number_of_volumes, 2)

    self.assertEqual(volume_system.volume_identifiers, ['lvm1', 'lvm2'])

    volume = volume_system.GetVolumeByIndex(0)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 1)
    self.assertEqual(volume.identifier, 'lvm1')

    expected_value = 'EMfcGR-KAOf-VrvI-yqth-fmz0-Ylmx-xepAJq'
    volume_attribute = volume.GetAttribute('identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    volume = volume_system.GetVolumeByIndex(99)
    self.assertIsNone(volume)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Core Storage (CS) volume system."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.volume import cs_volume_system

from tests import test_lib as shared_test_lib


class CSVolumeSystemTest(shared_test_lib.BaseTestCase):
  """Tests the Core Storage (CS) volume system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['fvdetest.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)
    test_gpt_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_GPT, location='/p1',
        parent=test_qcow_path_spec)
    self._cs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/',
        parent=test_gpt_path_spec)

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = cs_volume_system.CSVolumeSystem()

    volume_system.Open(self._cs_path_spec)

    self.assertEqual(volume_system.number_of_sections, 0)
    self.assertEqual(volume_system.number_of_volumes, 1)

    self.assertEqual(volume_system.volume_identifiers, ['cs1'])

    volume = volume_system.GetVolumeByIndex(0)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 1)
    self.assertEqual(volume.identifier, 'cs1')

    expected_value = '420af122-cf73-4a30-8b0a-a593a65fbef5'
    volume_attribute = volume.GetAttribute('identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, expected_value)

    volume = volume_system.GetVolumeByIndex(99)
    self.assertIsNone(volume)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Virtual File System (VFS) mount point manager."""

import platform
import unittest

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.mount import manager
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.resolver import resolver

from tests import test_lib as shared_test_lib


class MountPointManagerTest(shared_test_lib.BaseTestCase):
  """Class to test the mount point manager."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['ext2.qcow2'])
    self._SkipIfPathNotExists(test_path)

    self._resolver_context = context.Context()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)

  def testGetMountPoint(self):
    """Function to test the get mount point function."""
    manager.MountPointManager.RegisterMountPoint('C', self._qcow_path_spec)

    try:
      mount_point_path_spec = manager.MountPointManager.GetMountPoint('C')
      self.assertEqual(mount_point_path_spec, self._qcow_path_spec)

      mount_point_path_spec = manager.MountPointManager.GetMountPoint('D')
      self.assertIsNone(mount_point_path_spec)

    finally:
      manager.MountPointManager.DeregisterMountPoint('C')

  def testOpenFileObjectOnDirectory(self):
    """Function to test mount point resolving on a directory."""
    test_path = self._GetTestFilePath(['testdir_os'])
    self._SkipIfPathNotExists(test_path)

    if platform.system() == 'Windows':
      test_mounted_location = '\\file1.txt'
    else:
      test_mounted_location = '/file1.txt'

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)

    manager.MountPointManager.RegisterMountPoint(
        'testdir_os', test_os_path_spec)

    try:
      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='testdir_os')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_OS, location=test_mounted_location,
          parent=parent_path_spec)
      file_object = resolver.Resolver.OpenFileObject(
          path_spec, resolver_context=self._resolver_context)

      self.assertIsNotNone(file_object)
      self.assertEqual(file_object.get_size(), 6)

    finally:
      manager.MountPointManager.DeregisterMountPoint('testdir_os')

  def testOpenFileObjectOnImage(self):
    """Function to test mount point resolving on a storage media image."""
    manager.MountPointManager.RegisterMountPoint('C', self._qcow_path_spec)

    try:
      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='C')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.PREFERRED_EXT_BACK_END, location='/passwords.txt',
          parent=parent_path_spec)
      file_object = resolver.Resolver.OpenFileObject(
          path_spec, resolver_context=self._resolver_context)

      self.assertIsNotNone(file_object)
      self.assertEqual(file_object.get_size(), 116)

      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='D')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.PREFERRED_EXT_BACK_END, location='/passwords.txt',
          parent=parent_path_spec)

      with self.assertRaises(errors.MountPointError):
        file_object = resolver.Resolver.OpenFileObject(
            path_spec, resolver_context=self._resolver_context)

    finally:
      manager.MountPointManager.DeregisterMountPoint('C')


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The Core Storage (CS) credentials."""

from dfvfs.credentials import credentials
from dfvfs.credentials import manager
from dfvfs.lib import definitions


class CSCredentials(credentials.Credentials):
  """Core Storage (CS) credentials."""

  # TODO: add support for key_data credential, needs pyfvde update.
  CREDENTIALS = frozenset([
      'encrypted_root_plist', 'password', 'recovery_password'])

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_CS


manager.CredentialsManager.RegisterCredentials(CSCredentials())



# -*- coding: utf-8 -*-
"""The format specification classes."""


class Signature(object):
  """Signature of a format specification.

  The signature consists of a byte string pattern, an optional
  offset relative to the start of the data, and a value to indicate
  if the pattern is bound to the offset.

  Attributes:
    identifier (str): unique signature identifier for a specification store.
    offset (int): offset of the signature, where None indicates the signature
          has no offset. A positive offset or 0 is relative to the start of
          the data a negative offset is relative to the end of the data.
    pattern (bytes): pattern of the signature.
  """

  def __init__(self, pattern, offset=None):
    """Initializes a signature.

    Args:
      pattern (bytes): pattern of the signature. Wildcards or regular pattern
          (regexp) are not supported.
      offset (Optional[int]): offset of the signature, where None indicates
          the signature has no offset. A positive offset or 0 is relative
          from the start of the data a negative offset is relative to
          the end of the data.
    """
    super(Signature, self).__init__()
    self.identifier = None
    self.offset = offset
    self.pattern = pattern

  def SetIdentifier(self, identifier):
    """Sets the identifier of the signature in the specification store.

    Args:
      identifier (str): unique signature identifier for a specification store.
    """
    self.identifier = identifier


class FormatSpecification(object):
  """Format specification."""

  def __init__(self, identifier):
    """Initializes a specification.

    Args:
      identifier (str): unique name for the format.
    """
    super(FormatSpecification, self).__init__()
    self.identifier = identifier
    self.signatures = []

  def AddNewSignature(self, pattern, offset=None):
    """Adds a signature.

    Args:
      pattern (bytes): pattern of the signature. Wildcards or regular pattern
          (regexp) are not supported.
      offset (Optional[int]): offset of the signature, where None indicates
          the signature has no offset. A positive offset or 0 is relative
          from the start of the data a negative offset is relative to
          the end of the data.
    """
    self.signatures.append(Signature(pattern, offset=offset))


class FormatSpecificationStore(object):
  """Store for format specifications."""

  def __init__(self):
    """Initializes a format specification store."""
    super(FormatSpecificationStore, self).__init__()
    self._format_specifications = {}
    # Maps signature identifiers to format specifications.
    self._signature_map = {}

  @property
  def specifications(self):
    """generator[FormatSpecification]: format specifications."""
    return iter(self._format_specifications.values())

  def AddNewSpecification(self, identifier):
    """Adds a new format specification.

    Args:
      identifier (str): unique signature identifier for a specification store.

    Returns:
      FormatSpecification: format specification.

    Raises:
      KeyError: if the store already contains a specification with
          the same identifier.
    """
    if identifier in self._format_specifications:
      raise KeyError(
          f'Format specification {identifier:s} is already defined in store.')

    self._format_specifications[identifier] = FormatSpecification(identifier)

    return self._format_specifications[identifier]

  def AddSpecification(self, specification):
    """Adds a specification.

    Args:
      specification (FormatSpecification): format specification.

    Raises:
      KeyError: if the store already contains a specification with
          the same identifier.
    """
    if specification.identifier in self._format_specifications:
      raise KeyError((
          f'Format specification {specification.identifier:s} is already '
          f'defined in store.'))

    self._format_specifications[specification.identifier] = specification

    for signature in specification.signatures:
      signature_index = len(self._signature_map)

      signature_identifier = f'{specification.identifier:s}:{signature_index:d}'
      if signature_identifier in self._signature_map:
        raise KeyError(
            f'Signature {signature_identifier:s} is already defined in map.')

      signature.SetIdentifier(signature_identifier)
      self._signature_map[signature_identifier] = specification

  def GetSpecificationBySignature(self, signature_identifier):
    """Retrieves a specification mapped to a signature identifier.

    Args:
      signature_identifier (str): unique signature identifier for
          a specification store.

    Returns:
      FormatSpecification: A format specification or None if the signature
          identifier does not exist within the specification store.
    """
    return self._signature_map.get(signature_identifier, None)



# -*- coding: utf-8 -*-
"""The compressed stream file system implementation."""

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import compressed_stream_path_spec
from dfvfs.vfs import compressed_stream_file_entry
from dfvfs.vfs import root_only_file_system


class CompressedStreamFileSystem(root_only_file_system.RootOnlyFileSystem):
  """Compressed stream file system."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_COMPRESSED_STREAM

  def __init__(self, resolver_context, path_spec):
    """Initializes a compressed stream file system.

    Args:
      resolver_context (Context): a resolver context.
      path_spec (PathSpec): a path specification.
    """
    super(CompressedStreamFileSystem, self).__init__(
        resolver_context, path_spec)
    self._compression_method = None

  def _Close(self):
    """Closes the file system.

    Raises:
      IOError: if the close failed.
    """
    self._compression_method = None

  def _Open(self, mode='rb'):
    """Opens the file system defined by path specification.

    Args:
      mode (Optional[str]): file access mode. The default is 'rb' which
          represents read-only binary.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file system could not be opened.
      PathSpecError: if the path specification is incorrect.
      ValueError: if the path specification is invalid.
    """
    if not self._path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification without parent.')

    compression_method = getattr(self._path_spec, 'compression_method', None)
    if not compression_method:
      raise errors.PathSpecError(
          'Unsupported path specification without compression method.')

    self._compression_method = compression_method

  def GetFileEntryByPathSpec(self, path_spec):
    """Retrieves a file entry for a path specification.

    Args:
      path_spec (PathSpec): a path specification.

    Returns:
      CompressedStreamFileEntry: a file entry or None if not available.
    """
    return compressed_stream_file_entry.CompressedStreamFileEntry(
        self._resolver_context, self, path_spec, is_root=True, is_virtual=True)

  def GetRootFileEntry(self):
    """Retrieves the root file entry.

    Returns:
      CompressedStreamFileEntry: a file entry or None if not available.
    """
    path_spec = compressed_stream_path_spec.CompressedStreamPathSpec(
        compression_method=self._compression_method,
        parent=self._path_spec.parent)
    return self.GetFileEntryByPathSpec(path_spec)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the VFS data stream interface."""

import unittest

from dfvfs.vfs import data_stream

from tests import test_lib as shared_test_lib


class DataStreamTest(shared_test_lib.BaseTestCase):
  """Tests the VFS data stream interface."""

  def testName(self):
    """Test the name property."""
    test_data_stream = data_stream.DataStream(None)
    self.assertEqual(test_data_stream.name, '')

  def testGetExtents(self):
    """Test the GetExtents function."""
    test_data_stream = data_stream.DataStream(None)
    extents = test_data_stream.GetExtents()
    self.assertEqual(extents, [])

  def testIsDefault(self):
    """Test the IsDefault function."""
    test_data_stream = data_stream.DataStream(None)
    result = test_data_stream.IsDefault()
    self.assertTrue(result)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The APFS container directory implementation."""

from dfvfs.path import apfs_container_path_spec
from dfvfs.vfs import directory


class APFSContainerDirectory(directory.Directory):
  """File system directory that uses pyfsapfs."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      APFSContainerPathSpec: a path specification.
    """
    # Only the virtual root file has directory entries.
    volume_index = self._file_system.GetVolumeIndexByPathSpec(self.path_spec)
    if volume_index is None:
      location = getattr(self.path_spec, 'location', None)
      if location and location == self._file_system.LOCATION_ROOT:
        fsapfs_container = self._file_system.GetAPFSContainer()

        for volume_index in range(0, fsapfs_container.number_of_volumes):
          apfs_volume_index = volume_index + 1
          apfs_volume_location = f'/apfs{apfs_volume_index:d}'
          yield apfs_container_path_spec.APFSContainerPathSpec(
              location=apfs_volume_location, parent=self.path_spec.parent,
              volume_index=volume_index)



# -*- coding: utf-8 -*-
"""The credentials interface."""


class Credentials(object):
  """Credentials interface."""

  def __init__(self):
    """Initializes credentials.

    Raises:
      ValueError: if a derived credentials class does not define a type
          indicator.
    """
    super(Credentials, self).__init__()

    if not getattr(self, 'TYPE_INDICATOR', None):
      raise ValueError('Missing type indicator.')

  @property
  def type_indicator(self):
    """str: type indicator."""
    # pylint: disable=no-member
    return self.TYPE_INDICATOR



# -*- coding: utf-8 -*-
"""The Volume Shadow Snapshots (VSS) file system implementation."""

import pyvshadow

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.lib import vshadow_helper
from dfvfs.path import vshadow_path_spec
from dfvfs.resolver import resolver
from dfvfs.vfs import file_system
from dfvfs.vfs import vshadow_file_entry


class VShadowFileSystem(file_system.FileSystem):
  """File system that uses pyvshadow."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_VSHADOW

  def __init__(self, resolver_context, path_spec):
    """Initializes a file system.

    Args:
      resolver_context (Context): resolver context.
      path_spec (PathSpec): a path specification.
    """
    super(VShadowFileSystem, self).__init__(resolver_context, path_spec)
    self._file_object = None
    self._vshadow_volume = None

  def _Close(self):
    """Closes the file system.

    Raises:
      IOError: if the close failed.
    """
    self._vshadow_volume.close()
    self._vshadow_volume = None
    self._file_object = None

  def _Open(self, mode='rb'):
    """Opens the file system object defined by path specification.

    Args:
      mode (Optional[str]): file access mode. The default is 'rb' which
          represents read-only binary.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file system object could not be opened.
      PathSpecError: if the path specification is incorrect.
      ValueError: if the path specification is invalid.
    """
    if not self._path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification without parent.')

    file_object = resolver.Resolver.OpenFileObject(
        self._path_spec.parent, resolver_context=self._resolver_context)

    vshadow_volume = pyvshadow.volume()
    vshadow_volume.open_file_object(file_object)

    self._file_object = file_object
    self._vshadow_volume = vshadow_volume

  def FileEntryExistsByPathSpec(self, path_spec):
    """Determines if a file entry for a path specification exists.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      bool: True if the file entry exists.
    """
    store_index = vshadow_helper.VShadowPathSpecGetStoreIndex(path_spec)

    # The virtual root file has no corresponding store index but
    # should have a location.
    if store_index is None:
      location = getattr(path_spec, 'location', None)
      return location is not None and location == self.LOCATION_ROOT

    return 0 <= store_index < self._vshadow_volume.number_of_stores

  def GetFileEntryByPathSpec(self, path_spec):
    """Retrieves a file entry for a path specification.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      VShadowFileEntry: file entry or None if not available.
    """
    store_index = vshadow_helper.VShadowPathSpecGetStoreIndex(path_spec)

    # The virtual root file has no corresponding store index but
    # should have a location.
    if store_index is None:
      location = getattr(path_spec, 'location', None)
      if location is None or location != self.LOCATION_ROOT:
        return None

      return vshadow_file_entry.VShadowFileEntry(
          self._resolver_context, self, path_spec, is_root=True,
          is_virtual=True)

    if store_index < 0 or store_index >= self._vshadow_volume.number_of_stores:
      return None

    return vshadow_file_entry.VShadowFileEntry(
        self._resolver_context, self, path_spec)

  def GetRootFileEntry(self):
    """Retrieves the root file entry.

    Returns:
      VShadowFileEntry: file entry or None if not available.
    """
    path_spec = vshadow_path_spec.VShadowPathSpec(
        location=self.LOCATION_ROOT, parent=self._path_spec.parent)
    return self.GetFileEntryByPathSpec(path_spec)

  def GetVShadowStoreByPathSpec(self, path_spec):
    """Retrieves a VSS store for a path specification.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      pyvshadow.store: a VSS store or None if not available.
    """
    store_index = vshadow_helper.VShadowPathSpecGetStoreIndex(path_spec)
    if store_index is None:
      return None

    return self._vshadow_volume.get_store(store_index)

  def GetVShadowVolume(self):
    """Retrieves a VSS volume.

    Returns:
      pyvshadow.volume: a VSS volume.
    """
    return self._vshadow_volume



# -*- coding: utf-8 -*-
"""The NTFS data stream implementation."""

from dfvfs.lib import definitions
from dfvfs.vfs import data_stream
from dfvfs.vfs import extent


class NTFSDataStream(data_stream.DataStream):
  """File system data stream that uses pyfsntfs."""

  def __init__(self, file_entry, fsntfs_data_stream):
    """Initializes a NTFS data stream.

    Args:
      file_entry (FileEntry): file entry.
      fsntfs_data_stream (pyfsntfs.data_stream): NTFS data stream.
    """
    super(NTFSDataStream, self).__init__(file_entry)
    self._fsntfs_data_stream = fsntfs_data_stream

    if fsntfs_data_stream:
      self._name = fsntfs_data_stream.name

  def GetExtents(self):
    """Retrieves the extents.

    Returns:
      list[Extent]: the extents of the data stream.
    """
    if not self._fsntfs_data_stream:
      return super(NTFSDataStream, self).GetExtents()

    extents = []
    for extent_index in range(self._fsntfs_data_stream.number_of_extents):
      extent_offset, extent_size, extent_flags = (
          self._fsntfs_data_stream.get_extent(extent_index))

      if extent_flags & 0x1:
        extent_type = definitions.EXTENT_TYPE_SPARSE
      else:
        extent_type = definitions.EXTENT_TYPE_DATA

      data_stream_extent = extent.Extent(
          extent_type=extent_type, offset=extent_offset, size=extent_size)
      extents.append(data_stream_extent)

    return extents



# -*- coding: utf-8 -*-
"""The FAT file entry implementation."""

import pyfsfat

from dfdatetime import definitions as dfdatetime_definitions
from dfdatetime import fat_date_time as dfdatetime_fat_date_time

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import fat_path_spec
from dfvfs.resolver import resolver
from dfvfs.vfs import attribute
from dfvfs.vfs import extent
from dfvfs.vfs import fat_directory
from dfvfs.vfs import file_entry


class FATFileEntry(file_entry.FileEntry):
  """File system file entry that uses pyfsfat."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_FAT

  _FILE_SYSTEM_FORMAT_EXFAT = pyfsfat.file_system_formats.EXFAT

  def __init__(
      self, resolver_context, file_system, path_spec, fsfat_file_entry=None,
      is_root=False, is_virtual=False):
    """Initializes a file entry.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsfat_file_entry (Optional[pyfsfat.file_entry]): FAT file entry.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file
          entry emulated by the corresponding file system.

    Raises:
      BackEndError: if the pyfsfat file entry is missing.
    """
    if not fsfat_file_entry:
      fsfat_file_entry = file_system.GetFATFileEntryByPathSpec(path_spec)
    if not fsfat_file_entry:
      raise errors.BackEndError('Missing pyfsfat file entry.')

    fsfat_volume = file_system.GetFATVolume()

    if is_root:
      file_entry_name = ''
    else:
      file_entry_name = fsfat_file_entry.name

    super(FATFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._file_system_format = fsfat_volume.get_file_system_format()
    self._fsfat_file_entry = fsfat_file_entry
    self._name = file_entry_name

    file_attribute_flags = fsfat_file_entry.file_attribute_flags
    if (file_attribute_flags is None or
        file_attribute_flags & pyfsfat.file_attribute_flags.DIRECTORY):
      self.entry_type = definitions.FILE_ENTRY_TYPE_DIRECTORY
    elif file_attribute_flags & pyfsfat.file_attribute_flags.DEVICE:
      self.entry_type = definitions.FILE_ENTRY_TYPE_DEVICE
    else:
      self.entry_type = definitions.FILE_ENTRY_TYPE_FILE

  def _GetDirectory(self):
    """Retrieves a directory.

    Returns:
      FATDirectory: directory or None if not available.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_DIRECTORY:
      return None

    return fat_directory.FATDirectory(
        self._file_system, self.path_spec, self._fsfat_file_entry)

  def _GetStatAttribute(self):
    """Retrieves a stat attribute.

    Returns:
      StatAttribute: a stat attribute or None if not available.
    """
    stat_attribute = attribute.StatAttribute()
    stat_attribute.inode_number = self._fsfat_file_entry.identifier
    stat_attribute.size = self._fsfat_file_entry.size
    stat_attribute.type = self.entry_type

    return stat_attribute

  def _GetSubFileEntries(self):
    """Retrieves a sub file entries generator.

    Yields:
      FATFileEntry: a sub file entry.
    """
    if self._directory is None:
      self._directory = self._GetDirectory()

    if self._directory:
      for path_spec in self._directory.entries:
        yield FATFileEntry(self._resolver_context, self._file_system, path_spec)

  @property
  def access_time(self):
    """dfdatetime.DateTimeValues: access time or None if not available."""
    timestamp = self._fsfat_file_entry.get_access_time_as_integer()
    # Access time can be None if not present and 0 if not set.
    if not timestamp:
      return None

    if self._file_system_format == self._FILE_SYSTEM_FORMAT_EXFAT:
      precision = dfdatetime_definitions.PRECISION_10_MILLISECONDS
    else:
      precision = dfdatetime_definitions.PRECISION_1_DAY

    return dfdatetime_fat_date_time.FATTimestamp(
        precision=precision, timestamp=timestamp)

  @property
  def creation_time(self):
    """dfdatetime.DateTimeValues: creation time or None if not available."""
    timestamp = self._fsfat_file_entry.get_creation_time_as_integer()
    # Creation time can be None if not present and 0 if not set.
    if not timestamp:
      return None

    return dfdatetime_fat_date_time.FATTimestamp(
        precision=dfdatetime_definitions.PRECISION_10_MILLISECONDS,
        timestamp=timestamp)

  @property
  def name(self):
    """str: name of the file entry, which does not include the full path."""
    return self._name

  @property
  def modification_time(self):
    """dfdatetime.DateTimeValues: modification time or None if not available."""
    timestamp = self._fsfat_file_entry.get_modification_time_as_integer()
    # Modification time can be None if not present and 0 if not set.
    if not timestamp:
      return None

    if self._file_system_format == self._FILE_SYSTEM_FORMAT_EXFAT:
      precision = dfdatetime_definitions.PRECISION_10_MILLISECONDS
    else:
      precision = dfdatetime_definitions.PRECISION_2_SECONDS

    return dfdatetime_fat_date_time.FATTimestamp(
        precision=precision, timestamp=timestamp)

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    return self._fsfat_file_entry.size

  def GetExtents(self):
    """Retrieves the extents.

    Returns:
      list[Extent]: the extents.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_FILE:
      return []

    extents = []
    for extent_index in range(self._fsfat_file_entry.number_of_extents):
      extent_offset, extent_size, extent_flags = (
          self._fsfat_file_entry.get_extent(extent_index))

      if extent_flags & 0x1:
        extent_type = definitions.EXTENT_TYPE_SPARSE
      else:
        extent_type = definitions.EXTENT_TYPE_DATA

      data_stream_extent = extent.Extent(
          extent_type=extent_type, offset=extent_offset, size=extent_size)
      extents.append(data_stream_extent)

    return extents

  def GetFileObject(self, data_stream_name=''):
    """Retrieves a file-like object of a specific data stream.

    Args:
      data_stream_name (Optional[str]): name of the data stream, where an empty
          string represents the default data stream.

    Returns:
      FileIO: a file-like object or None if not available.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_FILE or data_stream_name:
      return None

    return resolver.Resolver.OpenFileObject(
        self.path_spec, resolver_context=self._resolver_context)

  def GetFATFileEntry(self):
    """Retrieves the FAT file entry.

    Returns:
      pyfsfat.file_entry: FAT file entry.
    """
    return self._fsfat_file_entry

  def GetParentFileEntry(self):
    """Retrieves the parent file entry.

    Returns:
      FATFileEntry: parent file entry or None if not available.
    """
    parent_location = None

    location = getattr(self.path_spec, 'location', None)
    if location is not None:
      parent_location = self._file_system.DirnamePath(location)
      if parent_location == '':
        parent_location = self._file_system.PATH_SEPARATOR

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    path_spec = fat_path_spec.FATPathSpec(
        location=parent_location, parent=parent_path_spec)

    is_root = bool(parent_location == self._file_system.LOCATION_ROOT)

    return FATFileEntry(
        self._resolver_context, self._file_system, path_spec, is_root=is_root)



# -*- coding: utf-8 -*-
"""The LUKS Drive Encryption file-like object."""

import pyluksde

from dfvfs.file_io import file_object_io
from dfvfs.lib import luksde_helper
from dfvfs.lib import errors
from dfvfs.resolver import resolver


class LUKSDEFile(file_object_io.FileObjectIO):
  """File input/output (IO) object using pyluksde."""

  def _OpenFileObject(self, path_spec):
    """Opens the file-like object defined by path specification.

    Args:
      path_spec (PathSpec): path specification.

    Returns:
      pyvde.volume: LUKSDE volume file-like object.

    Raises:
      PathSpecError: if the path specification is incorrect.
    """
    if not path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification without parent.')

    resolver.Resolver.key_chain.ExtractCredentialsFromPathSpec(path_spec)

    file_object = resolver.Resolver.OpenFileObject(
        path_spec.parent, resolver_context=self._resolver_context)
    luksde_volume = pyluksde.volume()

    luksde_helper.LUKSDEOpenVolume(
        luksde_volume, path_spec, file_object, resolver.Resolver.key_chain)
    return luksde_volume

  @property
  def is_locked(self):
    """bool: True if the volume is locked."""
    return self._file_object.is_locked()



# -*- coding: utf-8 -*-
"""The decrypter interface."""

import abc


class Decrypter(object):
  """Decrypter interface."""

  def __init__(self, **kwargs):
    """Initializes a decrypter.

    Args:
      kwargs (dict): keyword arguments depending on the decrypter.

    Raises:
      ValueError: when there are unused keyword arguments.
    """
    if kwargs:
      keyword_arguments = ', '.join(kwargs)
      raise ValueError(f'Unused keyword arguments: {keyword_arguments:s}.')

    super(Decrypter, self).__init__()

  # pylint: disable=redundant-returns-doc
  @abc.abstractmethod
  def Decrypt(self, encrypted_data, finalize=False):
    """Decrypts the encrypted data.

    Args:
      encrypted_data (bytes): encrypted data.
      finalize (Optional[bool]): True if the end of data has been reached and
          the cipher context should be finalized.

    Returns:
      tuple[bytes, bytes]: decrypted data and remaining encrypted data.
    """



# -*- coding: utf-8 -*-
"""The Virtual File System (VFS) extent."""


class Extent(object):
  """Extent.

  Attributes:
    extent_type (str): type of the extent, for example EXTENT_TYPE_SPARSE.
    offset (int): offset of the extent relative from the start of the file
        system in bytes.
    size (int): size of the extent in bytes.
  """

  def __init__(self, extent_type=None, offset=None, size=None):
    """Initializes an extent.

    Args:
      extent_type (Optional[str]): type of the extent, for example
          EXTENT_TYPE_SPARSE.
      offset (Optional[int]): offset of the extent relative from the start of
          the file system in bytes.
      size (Optional{int]): size of the extent in bytes.
    """
    super(Extent, self).__init__()
    self.extent_type = extent_type
    self.offset = offset
    self.size = size



# -*- coding: utf-8 -*-
"""The EXT directory implementation."""

from dfvfs.path import ext_path_spec
from dfvfs.vfs import directory


class EXTDirectory(directory.Directory):
  """File system directory that uses pyfsext."""

  def __init__(self, file_system, path_spec, fsext_file_entry):
    """Initializes a directory.

    Args:
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsext_file_entry (pyfsext.file_entry): EXT file entry.
    """
    super(EXTDirectory, self).__init__(file_system, path_spec)
    self._fsext_file_entry = fsext_file_entry

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      EXTPathSpec: EXT path specification.
    """
    location = getattr(self.path_spec, 'location', None)

    for fsext_sub_file_entry in self._fsext_file_entry.sub_file_entries:
      directory_entry = fsext_sub_file_entry.name

      if not location or location == self._file_system.PATH_SEPARATOR:
        directory_entry = self._file_system.JoinPath([directory_entry])
      else:
        directory_entry = self._file_system.JoinPath([
            location, directory_entry])

      yield ext_path_spec.EXTPathSpec(
          inode=fsext_sub_file_entry.inode_number, location=directory_entry,
          parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""The NTFS attribute implementations."""

import pyfwnt

from dfdatetime import filetime as dfdatetime_filetime

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.vfs import attribute


class NTFSAttribute(attribute.Attribute):
  """File system attribute that uses pyfsntfs."""

  def __init__(self, fsntfs_attribute):
    """Initializes an attribute.

    Args:
      fsntfs_attribute (pyfsntfs.attribute): NTFS attribute.

    Raises:
      BackEndError: if the pyfsntfs attribute is missing.
    """
    if not fsntfs_attribute:
      raise errors.BackEndError('Missing pyfsntfs attribute.')

    super(NTFSAttribute, self).__init__()
    self._fsntfs_attribute = fsntfs_attribute

  @property
  def attribute_type(self):
    """The attribute type."""
    return self._fsntfs_attribute.attribute_type


class FileNameNTFSAttribute(NTFSAttribute):
  """NTFS $FILE_NAME file system attribute."""

  TYPE_INDICATOR = definitions.ATTRIBUTE_TYPE_NTFS_FILE_NAME

  @property
  def access_time(self):
    """dfdatetime.Filetime: access time or None if not set."""
    timestamp = self._fsntfs_attribute.get_access_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def creation_time(self):
    """dfdatetime.Filetime: creation time or None if not set."""
    timestamp = self._fsntfs_attribute.get_creation_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def entry_modification_time(self):
    """dfdatetime.Filetime: entry modification time or None if not set."""
    timestamp = self._fsntfs_attribute.get_entry_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def file_attribute_flags(self):
    """int: file attribute flags or None if not available."""
    return self._fsntfs_attribute.file_attribute_flags

  @property
  def modification_time(self):
    """dfdatetime.Filetime: modification time."""
    timestamp = self._fsntfs_attribute.get_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def name(self):
    """str: name."""
    return self._fsntfs_attribute.name

  @property
  def name_space(self):
    """int: name_space."""
    return self._fsntfs_attribute.name_space

  @property
  def parent_file_reference(self):
    """int: parent file reference."""
    return self._fsntfs_attribute.parent_file_reference


class ObjectIdentifierNTFSAttribute(NTFSAttribute):
  """NTFS $OBJECT_ID file system attribute."""

  TYPE_INDICATOR = definitions.ATTRIBUTE_TYPE_NTFS_OBJECT_ID

  @property
  def droid_file_identifier(self):
    """str: droid file identifier, formatted as an UUID."""
    return self._fsntfs_attribute.droid_file_identifier


class SecurityDescriptorNTFSAttribute(NTFSAttribute):
  """NTFS $SECURITY_DESCRIPTOR file system attribute."""

  TYPE_INDICATOR = definitions.ATTRIBUTE_TYPE_NTFS_SECURITY_DESCRIPTOR

  @property
  def security_descriptor(self):
    """pyfwnt.security_descriptor: security descriptor."""
    fwnt_security_descriptor = pyfwnt.security_descriptor()
    fwnt_security_descriptor.copy_from_byte_stream(self._fsntfs_attribute.data)
    return fwnt_security_descriptor


class StandardInformationNTFSAttribute(NTFSAttribute):
  """NTFS $STANDARD_INFORMATION file system attribute."""

  TYPE_INDICATOR = definitions.ATTRIBUTE_TYPE_NTFS_STANDARD_INFORMATION

  @property
  def access_time(self):
    """dfdatetime.Filetime: access time or None if not set."""
    timestamp = self._fsntfs_attribute.get_access_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def creation_time(self):
    """dfdatetime.Filetime: creation time or None if not set."""
    timestamp = self._fsntfs_attribute.get_creation_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def entry_modification_time(self):
    """dfdatetime.Filetime: entry modification time or None if not set."""
    timestamp = self._fsntfs_attribute.get_entry_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def file_attribute_flags(self):
    """int: file attribute flags or None if not available."""
    return self._fsntfs_attribute.file_attribute_flags

  @property
  def modification_time(self):
    """dfdatetime.Filetime: modification time or None if not set."""
    timestamp = self._fsntfs_attribute.get_modification_time_as_integer()
    return dfdatetime_filetime.Filetime(timestamp=timestamp)

  @property
  def owner_identifier(self):
    """int: owner identifier."""
    return self._fsntfs_attribute.owner_identifier

  @property
  def security_descriptor_identifier(self):
    """int: security descriptor identifier."""
    return self._fsntfs_attribute.security_descriptor_identifier

  @property
  def update_sequence_number(self):
    """int: update sequence number."""
    return self._fsntfs_attribute.update_sequence_number



# -*- coding: utf-8 -*-
"""The SQLite blob directory implementation."""

from dfvfs.path import sqlite_blob_path_spec
from dfvfs.vfs import directory


class SQLiteBlobDirectory(directory.Directory):
  """SQLite blob directory."""

  def __init__(self, file_system, path_spec):
    """Initializes a directory.

    Args:
      file_system (SQLiteBlobFileSystem): file system.
      path_spec (SQLiteBlobPathSpec): path specification.
    """
    super(SQLiteBlobDirectory, self).__init__(file_system, path_spec)
    self._number_of_entries = None

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      SQLiteBlobPathSpec: a path specification.
    """
    table_name = getattr(self.path_spec, 'table_name', None)
    column_name = getattr(self.path_spec, 'column_name', None)

    if table_name and column_name:
      if self._number_of_entries is None:
        # Open the first entry to determine how many entries we have.
        # TODO: change this when there is a move this to a central temp file
        # manager. https://github.com/log2timeline/dfvfs/issues/92
        path_spec = sqlite_blob_path_spec.SQLiteBlobPathSpec(
            table_name=table_name, column_name=column_name, row_index=0,
            parent=self.path_spec.parent)

        sub_file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
        if not sub_file_entry:
          self._number_of_entries = 0
        else:
          self._number_of_entries = sub_file_entry.GetNumberOfRows()

      for row_index in range(0, self._number_of_entries):
        yield sqlite_blob_path_spec.SQLiteBlobPathSpec(
            table_name=table_name, column_name=column_name, row_index=row_index,
            parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""The Core Storage (CS) volume system."""

from dfvfs.lib import definitions
from dfvfs.volume import factory
from dfvfs.volume import volume_system


class CSVolume(volume_system.Volume):
  """Volume that uses pyfvde."""

  def __init__(self, file_entry):
    """Initializes a CS volume.

    Args:
      file_entry (CSFileEntry): a CS file entry.
    """
    super(CSVolume, self).__init__(file_entry.name)
    self._file_entry = file_entry

  def _Parse(self):
    """Extracts attributes and extents from the volume."""
    fvde_logical_volume = self._file_entry.GetFVDELogicalVolume()

    volume_attribute = volume_system.VolumeAttribute(
        'identifier', fvde_logical_volume.identifier)
    self._AddAttribute(volume_attribute)

    # TODO: implement in pyfvde
    # TODO: add support for creation time
    # TODO: add support for logical volume extents
    volume_extent = volume_system.VolumeExtent(0, fvde_logical_volume.size)
    self._extents.append(volume_extent)


class CSVolumeSystem(volume_system.VolumeSystem):
  """Volume system that uses pyfvde."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_CS

  VOLUME_IDENTIFIER_PREFIX = 'cs'

  def _Parse(self):
    """Extracts sections and volumes from the volume system."""
    root_file_entry = self._file_system.GetRootFileEntry()

    for sub_file_entry in root_file_entry.sub_file_entries:
      volume = CSVolume(sub_file_entry)
      self._AddVolume(volume)


factory.Factory.RegisterVolumeSystem(CSVolumeSystem)



# -*- coding: utf-8 -*-
"""The CPIO archive file system implementation."""

from dfvfs.lib import cpio
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import cpio_path_spec
from dfvfs.resolver import resolver
from dfvfs.vfs import cpio_file_entry
from dfvfs.vfs import file_system


class CPIOFileSystem(file_system.FileSystem):
  """CPIO archive file system."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_CPIO

  def __init__(self, resolver_context, path_spec, encoding='utf-8'):
    """Initializes a CPIO archive file system.

    Args:
      resolver_context (Context): resolver context.
      path_spec (PathSpec): a path specification.
      encoding (Optional[str]): file entry name encoding.
    """
    super(CPIOFileSystem, self).__init__(resolver_context, path_spec)
    self._cpio_archive_file = None
    self._file_object = None
    self.encoding = encoding

  def _Close(self):
    """Closes the file system.

    Raises:
      IOError: if the close failed.
    """
    self._cpio_archive_file.Close()
    self._cpio_archive_file = None
    self._file_object = None

  def _Open(self, mode='rb'):
    """Opens the file system defined by path specification.

    Args:
      mode (Optional[str]): file access mode. The default is 'rb' which
          represents read-only binary.

    Raises:
      AccessError: if the access to open the file was denied.
      IOError: if the file system could not be opened.
      PathSpecError: if the path specification is incorrect.
      ValueError: if the path specification is invalid.
    """
    if not self._path_spec.HasParent():
      raise errors.PathSpecError(
          'Unsupported path specification without parent.')

    file_object = resolver.Resolver.OpenFileObject(
        self._path_spec.parent, resolver_context=self._resolver_context)

    cpio_archive_file = cpio.CPIOArchiveFile()
    cpio_archive_file.Open(file_object)

    self._file_object = file_object
    self._cpio_archive_file = cpio_archive_file

  def FileEntryExistsByPathSpec(self, path_spec):
    """Determines if a file entry for a path specification exists.

    Args:
      path_spec (PathSpec): a path specification.

    Returns:
      bool: True if the file entry exists.
    """
    location = getattr(path_spec, 'location', None)
    if location is None or not location.startswith(self.LOCATION_ROOT):
      return False

    if len(location) == 1:
      return True

    return self._cpio_archive_file.FileEntryExistsByPath(location[1:])

  def GetCPIOArchiveFile(self):
    """Retrieves the CPIO archive file.

    Returns:
      CPIOArchiveFile: a CPIO archive file.
    """
    return self._cpio_archive_file

  def GetCPIOArchiveFileEntryByPathSpec(self, path_spec):
    """Retrieves the CPIO archive file entry for a path specification.

    Args:
      path_spec (PathSpec): a path specification.

    Returns:
      CPIOArchiveFileEntry: CPIO archive file entry or None if not available.

    Raises:
      PathSpecError: if the path specification is incorrect.
    """
    location = getattr(path_spec, 'location', None)
    if location is None:
      raise errors.PathSpecError('Path specification missing location.')

    if not location.startswith(self.LOCATION_ROOT):
      raise errors.PathSpecError('Invalid location in path specification.')

    if len(location) == 1:
      return None

    return self._cpio_archive_file.GetFileEntryByPath(location[1:])

  def GetFileEntryByPathSpec(self, path_spec):
    """Retrieves a file entry for a path specification.

    Args:
      path_spec (PathSpec): a path specification.

    Returns:
      CPIOFileEntry: a file entry or None if not available.
    """
    location = getattr(path_spec, 'location', None)

    if (location is None or
        not location.startswith(self.LOCATION_ROOT)):
      return None

    if len(location) == 1:
      return cpio_file_entry.CPIOFileEntry(
          self._resolver_context, self, path_spec, is_root=True,
          is_virtual=True)

    cpio_archive_file_entry = self._cpio_archive_file.GetFileEntryByPath(
        location[1:])
    if cpio_archive_file_entry is None:
      return None

    return cpio_file_entry.CPIOFileEntry(
        self._resolver_context, self, path_spec,
        cpio_archive_file_entry=cpio_archive_file_entry)

  def GetRootFileEntry(self):
    """Retrieves the root file entry.

    Returns:
      CPIOFileEntry: a file entry or None if not available.
    """
    path_spec = cpio_path_spec.CPIOPathSpec(
        location=self.LOCATION_ROOT, parent=self._path_spec.parent)
    return self.GetFileEntryByPathSpec(path_spec)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the file-like object implementation using pyfvde."""

import unittest

from dfvfs.file_io import hfs_file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import resolver

from tests.file_io import test_lib


class CSFileTest(test_lib.HFSImageFileTestCase):
  """Tests the Core Storage (CS) file-like object."""

  _IDENTIFIER_ANOTHER_FILE = 26
  _IDENTIFIER_PASSWORDS_TXT = 25

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CSFileTest, self).setUp()
    test_path = self._GetTestFilePath(['cs_single_volume.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._cs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._raw_path_spec,
        volume_index=0)

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_ANOTHER_FILE,
        location='/a_directory/another_file', parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


class CSFileWithKeyChainTest(test_lib.HFSImageFileTestCase):
  """Tests the Core Storage (CS) file-like object.

  The credentials are passed via the key chain.
  """

  _FVDE_PASSWORD = 'fvde-TEST'

  _IDENTIFIER_ANOTHER_FILE = 24
  _IDENTIFIER_PASSWORDS_TXT = 26

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CSFileWithKeyChainTest, self).setUp()
    test_path = self._GetTestFilePath(['fvdetest.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)
    test_tsk_partition_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK_PARTITION, location='/p1',
        parent=test_qcow_path_spec)
    self._cs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=test_tsk_partition_path_spec,
        volume_index=0)

    resolver.Resolver.key_chain.SetCredential(
        self._cs_path_spec, 'password', self._FVDE_PASSWORD)

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_ANOTHER_FILE,
        location='/a_directory/another_file', parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


class CSFileWithPathSpecCredentialsTest(test_lib.HFSImageFileTestCase):
  """Tests the Core Storage (CS) file-like object.

  The credentials are passed via the path specification.
  """
  _FVDE_PASSWORD = 'fvde-TEST'

  _IDENTIFIER_ANOTHER_FILE = 24
  _IDENTIFIER_PASSWORDS_TXT = 26

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CSFileWithPathSpecCredentialsTest, self).setUp()
    test_path = self._GetTestFilePath(['fvdetest.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)
    test_tsk_partition_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK_PARTITION, location='/p1',
        parent=test_qcow_path_spec)
    self._cs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=test_tsk_partition_path_spec,
        password=self._FVDE_PASSWORD, volume_index=0)

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_ANOTHER_FILE,
        location='/a_directory/another_file', parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_HFS,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='/passwords.txt',
        parent=self._cs_path_spec)
    file_object = hfs_file_io.HFSFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the encrypted stream file entry implementation."""

import unittest

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.resolver import resolver
from dfvfs.vfs import encrypted_stream_file_entry
from dfvfs.vfs import encrypted_stream_file_system

from tests import test_lib as shared_test_lib


class EncryptedStreamFileEntryTest(shared_test_lib.BaseTestCase):
  """Tests for the encrypted stream file entry."""

  _RC4_KEY = b'rc4test'

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.rc4'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._encrypted_stream_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_ENCRYPTED_STREAM,
        encryption_method=definitions.ENCRYPTION_METHOD_RC4,
        parent=test_os_path_spec)
    resolver.Resolver.key_chain.SetCredential(
        self._encrypted_stream_path_spec, 'key', self._RC4_KEY)

    self._file_system = encrypted_stream_file_system.EncryptedStreamFileSystem(
        self._resolver_context, self._encrypted_stream_path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testInitialize(self):
    """Test the __init__ function."""
    file_entry = encrypted_stream_file_entry.EncryptedStreamFileEntry(
        self._resolver_context, self._file_system,
        self._encrypted_stream_path_spec)
    self.assertIsNotNone(file_entry)

  def testSize(self):
    """Test the size property."""
    file_entry = encrypted_stream_file_entry.EncryptedStreamFileEntry(
        self._resolver_context, self._file_system,
        self._encrypted_stream_path_spec)

    self.assertIsNotNone(file_entry)

    try:
      self.assertEqual(file_entry.size, 1247)
    except errors.BackEndError:
      raise unittest.SkipTest('missing cryptograpy support')

  def testGetFileEntryByPathSpec(self):
    """Test the get a file entry by path specification functionality."""
    file_entry = self._file_system.GetFileEntryByPathSpec(
        self._encrypted_stream_path_spec)
    self.assertIsNotNone(file_entry)

  def testGetParentFileEntry(self):
    """Tests the GetParentFileEntry function."""
    file_entry = self._file_system.GetFileEntryByPathSpec(
        self._encrypted_stream_path_spec)
    self.assertIsNotNone(file_entry)

    parent_file_entry = file_entry.GetParentFileEntry()
    self.assertIsNone(parent_file_entry)

  def testIsFunctions(self):
    """Test the Is? functions."""
    file_entry = self._file_system.GetFileEntryByPathSpec(
        self._encrypted_stream_path_spec)
    self.assertIsNotNone(file_entry)

    self.assertTrue(file_entry.IsRoot())
    self.assertTrue(file_entry.IsVirtual())
    self.assertTrue(file_entry.IsAllocated())

    self.assertFalse(file_entry.IsDevice())
    self.assertFalse(file_entry.IsDirectory())
    self.assertTrue(file_entry.IsFile())
    self.assertFalse(file_entry.IsLink())
    self.assertFalse(file_entry.IsPipe())
    self.assertFalse(file_entry.IsSocket())

  def testSubFileEntries(self):
    """Test the sub file entries iteration functionality."""
    file_entry = self._file_system.GetFileEntryByPathSpec(
        self._encrypted_stream_path_spec)
    self.assertIsNotNone(file_entry)

    self.assertEqual(file_entry.number_of_sub_file_entries, 0)

    expected_sub_file_entry_names = []

    sub_file_entry_names = []
    for sub_file_entry in file_entry.sub_file_entries:
      sub_file_entry_names.append(sub_file_entry.name)

    self.assertEqual(
        len(sub_file_entry_names), len(expected_sub_file_entry_names))
    self.assertEqual(
        sorted(sub_file_entry_names), expected_sub_file_entry_names)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the GUID Partition Table (GPT) volume system."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.volume import gpt_volume_system

from tests import test_lib as shared_test_lib


class GPTVolumeSystemTest(shared_test_lib.BaseTestCase):
  """Tests the GUID Partition Table (GPT) volume system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['gpt.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._gpt_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_GPT, location='/',
        parent=test_raw_path_spec)

  # vsgptinfo gpt.raw
  #
  # GUID Partition Table (GPT) information:
  #   Disk identifier      : e86e657a-d840-4c09-afe3-a1a5f665cf44
  #   Bytes per sector     : 512
  #   Number of partitions : 2
  #
  # Partition: 1
  #   Identifier           : 1e25588c-27a9-4094-868c-2f257021f87b
  #   Type identifier      : 0fc63daf-8483-4772-8e79-3d69d8477de4
  #   Type                 : 0x00 (Empty)
  #   Offset               : 1048576 (0x00100000)
  #   Size                 : 65536
  #
  # Partition: 2
  #   Identifier           : 53d86ccf-3188-4b54-90d8-81866426b70a
  #   Type identifier      : 0fc63daf-8483-4772-8e79-3d69d8477de4
  #   Type                 : 0x00 (Empty)
  #   Offset               : 2097152 (0x00200000)
  #   Size                 : 65536

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = gpt_volume_system.GPTVolumeSystem()

    volume_system.Open(self._gpt_path_spec)

    self.assertEqual(volume_system.number_of_sections, 0)
    self.assertEqual(volume_system.number_of_volumes, 2)

    self.assertEqual(volume_system.volume_identifiers, ['p1', 'p2'])

    volume = volume_system.GetVolumeByIndex(0)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 1)
    self.assertEqual(volume.identifier, 'p1')

    volume_attribute = volume.GetAttribute('identifier')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(
        volume_attribute.value, '1e25588c-27a9-4094-868c-2f257021f87b')

    volume_extent = volume.extents[0]
    self.assertIsNotNone(volume_extent)
    self.assertEqual(volume_extent.offset, 1048576)
    self.assertEqual(volume_extent.size, 65536)
    self.assertEqual(volume_extent.extent_type, volume_extent.EXTENT_TYPE_DATA)

    volume = volume_system.GetVolumeByIndex(99)
    self.assertIsNone(volume)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Blowfish decrypter object."""

import unittest

from dfvfs.encryption import blowfish_decrypter
from dfvfs.lib import definitions

from tests.encryption import test_lib


class BlowfishDecrypterTestCase(test_lib.DecrypterTestCase):
  """Tests for the Blowfish decrypter object."""

  _BLOWFISH_INITIALIZATION_VECTOR = b'This IV!'
  _BLOWFISH_KEY = b'This is a key123'

  def testInitialization(self):
    """Tests the initialization method."""
    # Test missing initialization vector with valid block cipher mode.
    blowfish_decrypter.BlowfishDecrypter(
        cipher_mode=definitions.ENCRYPTION_MODE_ECB, key=self._BLOWFISH_KEY)

    # Test missing arguments.
    with self.assertRaises(ValueError):
      blowfish_decrypter.BlowfishDecrypter()

    # Test unsupported block cipher mode.
    with self.assertRaises(ValueError):
      blowfish_decrypter.BlowfishDecrypter(
          cipher_mode='bogus', key=self._BLOWFISH_KEY)

    # Test missing initialization vector.
    with self.assertRaises(ValueError):
      blowfish_decrypter.BlowfishDecrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC, key=self._BLOWFISH_KEY)

    # Test incorrect key size.
    key = b'This is a key that is larger than the max key size of 448 bits.'
    with self.assertRaises(ValueError):
      blowfish_decrypter.BlowfishDecrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_ECB, key=key)

    # Test incorrect initialization vector type.
    with self.assertRaises(TypeError):
      blowfish_decrypter.BlowfishDecrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector='Wrong IV type', key=self._BLOWFISH_KEY)

    # Test incorrect initialization vector size.
    with self.assertRaises(ValueError):
      blowfish_decrypter.BlowfishDecrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector=b'Wrong IV size', key=self._BLOWFISH_KEY)

  def testDecrypt(self):
    """Tests the Decrypt method."""
    decrypter = blowfish_decrypter.BlowfishDecrypter(
        cipher_mode=definitions.ENCRYPTION_MODE_CBC,
        initialization_vector=self._BLOWFISH_INITIALIZATION_VECTOR,
        key=self._BLOWFISH_KEY)

    # Test full decryption.
    expected_decrypted_data = b'This is secret encrypted text!!!'

    decrypted_data, remaining_encrypted_data = decrypter.Decrypt(
        b'}\x00\x99\xd2\xab\x1c\xcd\x80y\xef\x0b\x0f\xf72Rp\xbb\\h\x06\xff\x07'
        b'\x9a\xcfE\r\x8d\x18\x90\x8e\xfe\xa3', finalize=True)

    self.assertEqual(decrypted_data, expected_decrypted_data)
    self.assertEqual(remaining_encrypted_data, b'')

    # Reset decrypter.
    decrypter = blowfish_decrypter.BlowfishDecrypter(
        cipher_mode=definitions.ENCRYPTION_MODE_CBC,
        initialization_vector=self._BLOWFISH_INITIALIZATION_VECTOR,
        key=self._BLOWFISH_KEY)

    # Test partial decryption.
    partial_encrypted_data = b'}\x00\x99\xd2\xab\x1c\xcd\x80y\xef'

    decrypted_data, remaining_encrypted_data = decrypter.Decrypt(
        partial_encrypted_data)

    self.assertEqual(decrypted_data, b'This is ')
    self.assertEqual(remaining_encrypted_data, b'y\xef')


if __name__ == '__main__':
  unittest.main()
